# TODO-future-barret; Maybe use a `._options()` method to return a `JsonifiableDict` within the DataGrid/DataTable classes?
# TODO-future-barret; Really feals as if we should have a base class that does most of this for us

from __future__ import annotations

import abc
from typing import Generic, Literal

from ..._docstring import add_example
from ._selection import (
    RowSelectionModeDeprecated,
    SelectionModeInput,
    SelectionModes,
    as_selection_modes,
)
from ._styles import StyleFn, StyleInfo, as_browser_style_infos, as_style_infos
from ._tbl_data import assert_data_is_not_none, serialize_frame
from ._types import FrameJson, IntoDataFrameT


class AbstractTabularData(abc.ABC):
    @abc.abstractmethod
    def to_payload(self) -> FrameJson: ...


@add_example(ex_dir="../../api-examples/data_frame_grid_table")
@add_example(ex_dir="../../api-examples/data_frame_styles")
class DataGrid(AbstractTabularData, Generic[IntoDataFrameT]):
    """
    Holds the data and options for a :class:`~shiny.render.data_frame` output, for a
    spreadsheet-like view.

    This class is used to wrap the returned data frame from a `@render.data_frame`
    render function. It allows you to specify options for the data grid, such as the
    width and height of the grid, whether to show a summary message, whether to show
    filter inputs, whether the cells are editable, and how the cells are selected.

    While there are currently no execution or parameter differences between `DataGrid`
    and `DataTable` other than CSS styling in the browser, the two classes are kept
    separate to allow for future extensibility.

    Parameters
    ----------
    data
        A [pandas](https://pandas.pydata.org/), [polars](https://pola.rs/), or
        eager [`narwhals`](https://narwhals-dev.github.io/narwhals/) compatible `DataFrame`
        object.
    width
        A _maximum_ amount of horizontal space for the data grid to occupy, in CSS units
        (e.g. `"400px"`) or as a number, which will be interpreted as pixels. The
        default is `fit-content`, which sets the grid's width according to its contents.
        Set this to `100%` to use the maximum available horizontal space.
    height
        A _maximum_ amount of vertical space for the data grid to occupy, in CSS units
        (e.g. `"400px"`) or as a number, which will be interpreted as pixels. If there
        are more rows than can fit in this space, the grid will scroll. Set the height
        to `"auto"` to allow the grid to grow to fit all of the rows (this is not
        recommended for large data sets, as it may crash the browser).
    summary
        If `True` (the default), shows a message like "Viewing rows 1 through 10 of 20"
        below the grid when not all of the rows are being shown. If `False`, the message
        is not displayed. You can also specify a string template to customize the
        message, containing `{start}`, `{end}`, and `{total}` tokens. For example:
        `"Viendo filas {start} a {end} de {total}"`.
    filters
        If `True`, shows a row of filter inputs below the headers, one for each column.
    editable
        If `True`, allows the user to edit the cells in the grid. When a cell is edited,
        the new value is sent to the server for processing. The server can then return
        a new value for the cell, which will be displayed in the grid.
    selection_mode
        Single string or a `set`/`list`/`tuple` of string values to define possible ways
        to select data within the data frame.

        Supported values:
        * Use `"none"` to disable any cell selections or editing.
        * Use `"row"` to allow a single row to be selected at a time.
        * Use `"rows"` to allow multiple rows to be selected by clicking on them
        individually.

        Resolution rules:
        * If `"none"` is supplied, all other values will be ignored.
        * If both `"row"` and `"rows"` are supplied, `"row"` will be dropped (supporting `"rows"`).
    styles
        A style info object, a list of style info objects, or a function that receives
        the (possibly updated) data frame and returns a list of style info objects. The
        style info objects can be used to apply CSS styles to the data frame. If
        `styles=None`, no styling will be applied.

        Style info object key/value description:
        * `location`: This value `"body"` and is not required.
        * `rows`: The row numbers to which the style should be applied. If `None`, the
            style will be applied to all rows.
        * `cols`: The column numbers to which the style should be applied. If `None`,
            the style will be applied to all columns.
        * `style`: A dictionary of CSS properties and values to apply to the selected
            rows and columns. Traditional _kebab-cased_ CSS property names (e.g.
            `background-color`) will work in addition to _camelCased_ CSS property names
            (e.g. `backgroundColor`).
        * `class`: A string of CSS class names to apply to the selected rows and columns.

        If both `style` and `class` are missing or `None`, nothing will be applied. If
        both `rows` and `cols` are missing or `None`, the style will be applied to the
        complete data frame.
    row_selection_mode
        Deprecated. Please use `selection_mode=` instead.

    Returns
    -------
    :
        An object suitable for being returned from a `@render.data_frame`-decorated
        output function.

    See Also
    --------
    * :func:`~shiny.ui.output_data_frame` - The UI placeholder for a data frame output.
    * :class:`~shiny.render.data_frame` - The `render` method for data frames.
    * :class:`~shiny.render.DataTable` - A more _tabular_ view of the data.
    """

    data: IntoDataFrameT
    width: str | float | None
    height: str | float | None
    summary: bool | str
    filters: bool
    editable: bool
    selection_modes: SelectionModes
    styles: list[StyleInfo] | StyleFn[IntoDataFrameT]

    def __init__(
        self,
        data: IntoDataFrameT,
        *,
        width: str | float | None = "fit-content",
        height: str | float | None = None,
        summary: bool | str = True,
        filters: bool = False,
        editable: bool = False,
        selection_mode: SelectionModeInput = "none",
        styles: StyleInfo | list[StyleInfo] | StyleFn[IntoDataFrameT] | None = None,
        row_selection_mode: RowSelectionModeDeprecated = "deprecated",
    ):
        assert_data_is_not_none(data)
        self.data = data

        self.width = width
        self.height = height
        self.summary = summary
        self.filters = filters
        self.editable = bool(editable)
        self.selection_modes = as_selection_modes(
            selection_mode,
            name="DataGrid",
            row_selection_mode=row_selection_mode,
        )
        self.styles = as_style_infos(styles)

    def to_payload(self) -> FrameJson:
        """
        Converts the `DataGrid` object to a payload dictionary.

        Returns
        -------
        :
            The payload dictionary representing the `DataGrid` object.
        """
        res: FrameJson = {
            **serialize_frame(self.data),
            "options": {
                "width": self.width,
                "height": self.height,
                "summary": self.summary,
                "filters": self.filters,
                "editable": self.editable,
                "style": "grid",
                "fill": self.height is None,
                "styles": as_browser_style_infos(
                    self.styles,
                    into_data=self.data,
                ),
            },
        }
        return res


@add_example(ex_dir="../../api-examples/data_frame_grid_table")
@add_example(ex_dir="../../api-examples/data_frame_styles")
class DataTable(AbstractTabularData, Generic[IntoDataFrameT]):
    """
    Holds the data and options for a :class:`~shiny.render.data_frame` output, for a
    spreadsheet-like view.

    This class is used to wrap the returned data frame from a `@render.data_frame`
    render function. It allows you to specify options for the data table, such as the
    width and height of the table, whether to show a summary message, whether to show
    filter inputs, whether the cells are editable, and how the cells are selected.

    While there are currently no execution or parameter differences between `DataGrid`
    and `DataTable` other than CSS styling in the browser, the two classes are kept
    separate to allow for future extensibility.

    Parameters
    ----------
    data
        A [pandas](https://pandas.pydata.org/), [polars](https://pola.rs/), or
        eager [`narwhals`](https://narwhals-dev.github.io/narwhals/) compatible `DataFrame`
        object.
    width
        A _maximum_ amount of vertical space for the data table to occupy, in CSS units
        (e.g. `"400px"`) or as a number, which will be interpreted as pixels. The
        default is `fit-content`, which sets the table's width according to its
        contents. Set this to `100%` to use the maximum available horizontal space.
    height
        A _maximum_ amount of vertical space for the data table to occupy, in CSS units
        (e.g. `"400px"`) or as a number, which will be interpreted as pixels. If there
        are more rows than can fit in this space, the table body will scroll. Set the
        height to `None` to allow the table to grow to fit all of the rows (this is not
        recommended for large data sets, as it may crash the browser).
    summary
        If `True` (the default), shows a message like "Viewing rows 1 through 10 of 20"
        below the grid when not all of the rows are being shown. If `False`, the message
        is not displayed. You can also specify a string template to customize the
        message, containing `{start}`, `{end}`, and `{total}` tokens. For example:
        `"Viendo filas {start} a {end} de {total}"`.
    filters
        If `True`, shows a row of filter inputs below the headers, one for each column.
    editable
        If `True`, allows the user to edit the cells in the grid. When a cell is edited,
        the new value is sent to the server for processing. The server can then return
        a new value for the cell, which will be displayed in the grid.
    selection_mode
        Single string or a `set`/`list`/`tuple` of string values to define possible ways
        to select data within the data frame.

        Supported values:
        * Use `"none"` to disable any cell selections or editing.
        * Use `"row"` to allow a single row to be selected at a time.
        * Use `"rows"` to allow multiple rows to be selected by clicking on them
        individually.

        Resolution rules:
        * If `"none"` is supplied, all other values will be ignored.
        * If both `"row"` and `"rows"` are supplied, `"row"` will be dropped (supporting `"rows"`).
    styles
        A style info object, a list of style info objects, or a function that receives
        the (possibly updated) data frame and returns a list of style info objects. The
        style info objects can be used to apply CSS styles to the data frame. If
        `styles=None`, no styling will be applied.

        Style info object key/value description:
        * `location`: This value `"body"` and is not required.
        * `rows`: The row numbers to which the style should be applied. If `None`, the
            style will be applied to all rows.
        * `cols`: The column numbers to which the style should be applied. If `None`,
            the style will be applied to all columns.
        * `style`: A dictionary of CSS properties and values to apply to the selected
            rows and columns. Traditional _kebab-cased_ CSS property names (e.g.
            `background-color`) will work in addition to _camelCased_ CSS property names
            (e.g. `backgroundColor`).
        * `class`: A string of CSS class names to apply to the selected rows and columns.

        If both `style` and `class` are missing or `None`, nothing will be applied. If
        both `rows` and `cols` are missing or `None`, the style will be applied to the
        complete data frame.
    row_selection_mode
        Deprecated. Please use `mode={row_selection_mode}_row` instead.

    Returns
    -------
    :
        An object suitable for being returned from a `@render.data_frame`-decorated
        output function.

    See Also
    --------
    * :func:`~shiny.ui.output_data_frame` - The UI placeholder for a data frame output.
    * :class:`~shiny.render.data_frame` - The `render` method for data frames.
    * :class:`~shiny.render.DataTable` - A more _grid_ view of the data.
    """

    data: IntoDataFrameT
    width: str | float | None
    height: str | float | None
    summary: bool | str
    filters: bool
    editable: bool
    selection_modes: SelectionModes
    styles: list[StyleInfo] | StyleFn[IntoDataFrameT]

    def __init__(
        self,
        data: IntoDataFrameT,
        *,
        width: str | float | None = "fit-content",
        height: str | float | None = "500px",
        summary: bool | str = True,
        filters: bool = False,
        editable: bool = False,
        selection_mode: SelectionModeInput = "none",
        styles: StyleInfo | list[StyleInfo] | StyleFn[IntoDataFrameT] | None = None,
        row_selection_mode: Literal["deprecated"] = "deprecated",
    ):
        assert_data_is_not_none(data)

        self.data = data

        self.width = width
        self.height = height
        self.summary = summary
        self.filters = filters
        self.editable = bool(editable)
        self.selection_modes = as_selection_modes(
            selection_mode,
            name="DataTable",
            row_selection_mode=row_selection_mode,
        )
        self.styles = as_style_infos(styles)

    def to_payload(self) -> FrameJson:
        """
        Converts the `DataTable` object to a payload dictionary.

        Returns
        -------
        :
            The payload dictionary representing the `DataTable` object.
        """
        res: FrameJson = {
            **serialize_frame(self.data),
            "options": {
                "width": self.width,
                "height": self.height,
                "summary": self.summary,
                "filters": self.filters,
                "editable": self.editable,
                "style": "table",
                "styles": as_browser_style_infos(
                    self.styles,
                    into_data=self.data,
                ),
            },
        }
        return res



from __future__ import annotations


from htmltools import (
    TagList,
    Tag,
    TagChild,
    TagAttrs,
    TagAttrValue,
    tags,
    HTML,
    head_content,
    a,
    br,
    code,
    div,
    em,
    h1,
    h2,
    h3,
    h4,
    h5,
    h6,
    hr,
    img,
    p,
    pre,
    span,
    strong,
)

from ...ui import (
    fill,
)

from ...ui import (
    busy_indicators,
)

from ...ui import (
    AccordionPanel,
    AnimationOptions,
    CardItem,
    Chat,
    ShowcaseLayout,
    Sidebar,
    SliderStepArg,
    SliderValueArg,
    ValueBoxTheme,
    brush_opts,
    click_opts,
    dblclick_opts,
    help_text,
    hover_opts,
    include_css,
    include_js,
    input_action_button,
    input_action_link,
    input_checkbox,
    input_checkbox_group,
    input_switch,
    input_radio_buttons,
    input_dark_mode,
    input_date,
    input_date_range,
    input_file,
    input_numeric,
    input_password,
    input_select,
    input_selectize,
    input_slider,
    bind_task_button,
    input_task_button,
    input_text,
    input_text_area,
    panel_title,
    insert_accordion_panel,
    remove_accordion_panel,
    update_accordion,
    update_accordion_panel,
    update_sidebar,
    update_action_button,
    update_action_link,
    update_checkbox,
    update_switch,
    update_checkbox_group,
    update_radio_buttons,
    update_dark_mode,
    update_date,
    update_date_range,
    update_numeric,
    update_select,
    update_selectize,
    update_slider,
    update_task_button,
    update_text,
    update_text_area,
    update_navs,
    update_tooltip,
    update_popover,
    insert_ui,
    remove_ui,
    markdown,
    modal_button,
    modal,
    modal_show,
    modal_remove,
    notification_show,
    notification_remove,
    nav_spacer,
    Progress,
    Theme,
    value_box_theme,
    js_eval,
)

from ._cm_components import (
    sidebar,
    layout_sidebar,
    layout_column_wrap,
    layout_columns,
    card,
    card_body,
    card_header,
    card_footer,
    accordion,
    accordion_panel,
    nav_panel,
    nav_control,
    nav_menu,
    navset_bar,
    navset_card_pill,
    navset_card_tab,
    navset_card_underline,
    navset_hidden,
    navset_pill,
    navset_pill_list,
    navset_tab,
    navset_underline,
    value_box,
    panel_well,
    panel_conditional,
    panel_fixed,
    panel_absolute,
    tooltip,
    popover,
)

from ._page import (
    page_opts,
)

from ._hold import (
    hold,
)

__all__ = (
    # Imports from htmltools
    "TagList",
    "Tag",
    "TagChild",
    "TagAttrs",
    "TagAttrValue",
    "tags",
    "HTML",
    "head_content",
    "a",
    "br",
    "code",
    "div",
    "em",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "hr",
    "img",
    "p",
    "pre",
    "span",
    "strong",
    "tags",
    # Submodules
    "busy_indicators",
    "fill",
    # Imports from ...ui
    "AccordionPanel",
    "AnimationOptions",
    "CardItem",
    "Chat",
    "ShowcaseLayout",
    "Sidebar",
    "SliderStepArg",
    "SliderValueArg",
    "ValueBoxTheme",
    "brush_opts",
    "click_opts",
    "dblclick_opts",
    "help_text",
    "hover_opts",
    "include_css",
    "include_js",
    "input_action_button",
    "input_action_link",
    "input_checkbox",
    "input_checkbox_group",
    "input_switch",
    "input_radio_buttons",
    "input_dark_mode",
    "input_date",
    "input_date_range",
    "input_file",
    "input_numeric",
    "input_password",
    "input_select",
    "input_selectize",
    "input_slider",
    "bind_task_button",
    "input_task_button",
    "input_text",
    "panel_title",
    "input_text_area",
    "insert_accordion_panel",
    "remove_accordion_panel",
    "update_accordion",
    "update_accordion_panel",
    "update_sidebar",
    "update_action_button",
    "update_action_link",
    "update_checkbox",
    "update_switch",
    "update_checkbox_group",
    "update_radio_buttons",
    "update_dark_mode",
    "update_date",
    "update_date_range",
    "update_numeric",
    "update_select",
    "update_selectize",
    "update_slider",
    "update_task_button",
    "update_text",
    "update_text_area",
    "update_navs",
    "update_tooltip",
    "update_popover",
    "insert_ui",
    "remove_ui",
    "markdown",
    "modal_button",
    "modal",
    "modal_show",
    "modal_remove",
    "notification_show",
    "notification_remove",
    "nav_spacer",
    "Progress",
    "Theme",
    "value_box_theme",
    # Imports from ._cm_components
    "sidebar",
    "layout_sidebar",
    "layout_column_wrap",
    "layout_columns",
    "card",
    "card_body",
    "card_header",
    "card_footer",
    "accordion",
    "accordion_panel",
    "nav_panel",
    "nav_control",
    "nav_menu",
    "navset_bar",
    "navset_card_pill",
    "navset_card_tab",
    "navset_card_underline",
    "navset_hidden",
    "navset_pill",
    "navset_pill_list",
    "navset_tab",
    "navset_underline",
    "value_box",
    "panel_well",
    "panel_conditional",
    "panel_fixed",
    "panel_absolute",
    "popover",
    "tooltip",
    # Imports from ._page
    "page_opts",
    # Imports from ._hold
    "hold",
    "js_eval",
)


# This is used for unit tests to verify that shiny.ui and shiny.express.ui stay in sync.
_known_missing = {
    # Items from shiny.ui that don't have a counterpart in shiny.express.ui
    "shiny.ui": (
        "column",  # Deprecated in favor of layout_columns
        "row",  # Deprecated in favor of layout_columns
        "page_bootstrap",
        "page_fixed",
        "page_sidebar",
        "page_fillable",
        "page_navbar",
        "page_fluid",
        "page_auto",
        "page_output",
        "showcase_bottom",
        "showcase_left_center",
        "showcase_top_right",
        # Outputs automatically placed by render functions
        "download_button",
        "download_link",
        "output_plot",
        "output_image",
        "output_text",
        "output_code",
        "output_text_verbatim",
        "output_table",
        "output_ui",
        "output_data_frame",
        # Chat knows how to render itself in express
        "chat_ui",
    ),
    # Items from shiny.express.ui that don't have a counterpart in shiny.ui
    "shiny.express.ui": (
        "page_opts",
        "hold",
    ),
}



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("sidebar")


def test_sidebar_position_and_open(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    left_sidebar = controller.Sidebar(page, "sidebar_left")
    output_txt_left = controller.OutputTextVerbatim(page, "state_left")
    left_sidebar.set(True)
    output_txt_left.expect_value("input.sidebar_left(): True")
    left_sidebar.expect_open(True)
    left_sidebar.set(False)
    output_txt_left.expect_value("input.sidebar_left(): False")
    left_sidebar.expect_text("Left sidebar content")
    left_sidebar.expect_handle(True)
    left_sidebar.expect_open(False)
    left_sidebar.loc_handle.click()
    left_sidebar.expect_open(True)
    output_txt_left.expect_value("input.sidebar_left(): True")

    right_sidebar = controller.Sidebar(page, "sidebar_right")
    output_txt_right = controller.OutputTextVerbatim(page, "state_right")
    right_sidebar.expect_text("Right sidebar content")
    output_txt_right.expect_value("input.sidebar_right(): True")
    right_sidebar.expect_handle(True)
    right_sidebar.expect_open(True)
    right_sidebar.loc_handle.click()
    right_sidebar.expect_open(False)
    output_txt_right.expect_value("input.sidebar_right(): False")

    closed_sidebar = controller.Sidebar(page, "sidebar_closed")
    output_txt_closed = controller.OutputTextVerbatim(page, "state_closed")
    output_txt_closed.expect_value("input.sidebar_closed(): False")
    closed_sidebar.expect_handle(True)
    closed_sidebar.expect_open(False)
    closed_sidebar.loc_handle.click()
    closed_sidebar.expect_text("Closed sidebar content")
    closed_sidebar.expect_open(True)
    output_txt_closed.expect_value("input.sidebar_closed(): True")

    always_sidebar = controller.Sidebar(page, "sidebar_always")
    output_txt_always = controller.OutputTextVerbatim(page, "state_always")
    always_sidebar.expect_text("Always sidebar content")
    output_txt_always.expect_value("input.sidebar_always(): True")
    # Handle is included but it should have `display: none`
    always_sidebar.expect_handle(True)

    expect(always_sidebar.loc_handle).to_have_css("display", "none")



from shiny.express import ui

with ui.tooltip(id="default_tooltip_auto"):
    ui.input_action_button("btn", "A button", class_="mt-3")
    "An auto message"

ui.br()
ui.br()

with ui.tooltip(id="default_tooltip_top", placement="top"):
    ui.input_action_button("btn2", "A button", class_="mt-3")
    "A top message"

ui.br()
ui.br()

with ui.tooltip(id="default_tooltip_right", placement="right"):
    ui.input_action_button("btn3", "A button", class_="mt-3")
    "A right message"


ui.br()
ui.br()

with ui.tooltip(id="default_tooltip_left", placement="left"):
    ui.input_action_button("btn4", "A button", class_="mt-3")
    "A left message"



from model_plots import (
    plot_accuracy_over_time,
    plot_feature_importance,
    plot_loss_over_time,
)

from shiny.express import render, ui

ui.page_opts(title="Model Dashboard")

ui.markdown("Using `ui.layout_columns()` for the layout.")


with ui.layout_columns(
    col_widths={"sm": (5, 7, 12)},
    # row_heights=(2, 3),
    # height="700px",
):
    with ui.card(full_screen=True):
        ui.card_header("Loss Over Time")

        @render.plot
        def loss_over_time():
            return plot_loss_over_time()

    with ui.card(full_screen=True):
        ui.card_header("Accuracy Over Time")

        @render.plot
        def accuracy_over_time():
            return plot_accuracy_over_time()

    with ui.card(full_screen=True):
        ui.card_header("Feature Importance")

        @render.plot
        def feature_importance():
            return plot_feature_importance()



# This is not a Shiny application; it is meant to be imported by shared_app.py.

from shiny import reactive, session

# Print this at the console to make it clear that shared.py is loaded just once per run
# of the app; each additional session does not result in this file loading again.
print("Loading shared.py!")

# This is a variable that can be used and shared across multiple sessions. This can be
# useful for large data, or values that are expensive to compute. It can also be useful
# for mutable objects that you want to share across sessions.
data = ["This", "is", "a", "list", "of", "words"]

# Any reactive objects should be created without a session context.
with session.session_context(None):
    # This reactive value can be used by multiple sessions; if it is invalidated (in
    # other words, if the value is changed), it will trigger invalidations in all of
    # those sessions.
    rv = reactive.value(50)



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_navset_bar_kitchensink(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    # Update the page size to be wider
    page.set_viewport_size({"width": 1500, "height": 800})

    navset_tab = controller.NavsetTab(page, "navsets_collection")
    navset_bar_fixed_top_position_selected = controller.NavsetBar(
        page,
        "navset_bar_fixed_top_position_selected",
    )
    navset_bar_header_footer_fixed_bottom_position = controller.NavsetBar(
        page,
        "navset_bar_header_footer_fixed_bottom_position",
    )
    navset_bar_with_sidebar_collapsible_bg_inverse = controller.NavsetBar(
        page,
        "navset_bar_with_sidebar_collapsible_bg_inverse",
    )
    navset_bar_collapsible_underline_fixed_gap = controller.NavsetBar(
        page,
        "navset_bar_collapsible_underline_fixed_gap",
    )

    navset_tab.set("fixed-top")
    navset_bar_fixed_top_position_selected._expect_content_text("Panel B content")
    navset_bar_fixed_top_position_selected.expect_position("fixed-top")
    navset_bar_fixed_top_position_selected.expect_value("B")

    navset_tab.set("fixed-bottom")
    navset_bar_header_footer_fixed_bottom_position._expect_content_text(
        "Panel A content"
    )
    navset_bar_header_footer_fixed_bottom_position.expect_position("fixed-bottom")
    navset_bar_header_footer_fixed_bottom_position.expect_value("A")

    navset_tab.set("sticky-top")
    navset_bar_with_sidebar_collapsible_bg_inverse._expect_content_text(
        "Panel A content"
    )
    navset_bar_with_sidebar_collapsible_bg_inverse.expect_position("sticky-top")
    navset_bar_with_sidebar_collapsible_bg_inverse.expect_inverse(value=True)
    navset_bar_with_sidebar_collapsible_bg_inverse.expect_bg("DodgerBlue")
    navset_bar_with_sidebar_collapsible_bg_inverse.expect_sidebar(True)
    navset_bar_with_sidebar_collapsible_bg_inverse.expect_fluid(True)

    navset_tab.set("fixed")
    navset_bar_collapsible_underline_fixed_gap._expect_content_text("Panel A content")
    navset_bar_collapsible_underline_fixed_gap.expect_value("A")
    navset_bar_collapsible_underline_fixed_gap.expect_gap("50px")
    navset_bar_collapsible_underline_fixed_gap.expect_fluid(False)



# pyright: reportUntypedFunctionDecorator=false
# pyright: reportUnknownMemberType=false
# pyright: reportInvalidTypeForm=false
# pyright: reportUnknownParameterType=false
# pyright: reportUnknownArgumentType=false
from __future__ import annotations

from typing import Optional, overload

from shiny import App, Inputs, Outputs, Session, ui
from shiny.render.transformer import (
    TransformerMetadata,
    ValueFn,
    is_async_callable,
    output_transformer,
    resolve_value_fn,
)


@output_transformer
async def TestTextTransformer(
    _meta: TransformerMetadata,
    _fn: ValueFn[str | None],
    *,
    extra_txt: Optional[str] = None,
) -> str | None:
    value = await resolve_value_fn(_fn)
    value = str(value)
    value += "; "
    value += "async" if is_async_callable(_fn) else "sync"
    if extra_txt:
        value = value + "; " + str(extra_txt)
    return value


@overload
def render_test_text(
    *, extra_txt: Optional[str] = None
) -> TestTextTransformer.OutputRendererDecorator: ...


@overload
def render_test_text(
    _fn: TestTextTransformer.ValueFn,
) -> TestTextTransformer.OutputRenderer: ...


def render_test_text(
    _fn: TestTextTransformer.ValueFn | None = None,
    *,
    extra_txt: Optional[str] = None,
) -> TestTextTransformer.OutputRenderer | TestTextTransformer.OutputRendererDecorator:
    return TestTextTransformer(
        _fn,
        TestTextTransformer.params(extra_txt=extra_txt),
    )


app_ui = ui.page_fluid(
    ui.code("t1:"),
    ui.output_text_verbatim("t1"),
    ui.code("t2:"),
    ui.output_text_verbatim("t2"),
    ui.code("t3:"),
    ui.output_text_verbatim("t3"),
    ui.code("t4:"),
    ui.output_text_verbatim("t4"),
    ui.code("t5:"),
    ui.output_text_verbatim("t5"),
    ui.code("t6:"),
    ui.output_text_verbatim("t6"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render_test_text
    def t1():
        return "t1; no call"
        # return "hello"

    @render_test_text
    async def t2():
        return "t2; no call"

    @render_test_text()
    def t3():
        return "t3; call"

    @render_test_text()
    async def t4():
        return "t4; call"

    @render_test_text(extra_txt="w/ extra_txt")
    def t5():
        return "t5; call"

    @render_test_text(extra_txt="w/ extra_txt")
    async def t6():
        return "t6; call"


app = App(app_ui, server)



# ------------------------------------------------------------------------------------
# A simple recipe extractor chatbot that extracts recipes from URLs using the OpenAI API.
# To run it, you'll need an OpenAI API key.
# To get one, follow the instructions at https://platform.openai.com/docs/quickstart
# ------------------------------------------------------------------------------------
import os

from openai import AsyncOpenAI
from utils import recipe_prompt, scrape_page_with_url

from shiny.express import ui

# Provide your API key here (or set the environment variable)
llm = AsyncOpenAI(api_key=os.environ.get("OPENAI_API_KEY"))

# Set some Shiny page options
ui.page_opts(
    title="Recipe Extractor Chat",
    fillable=True,
    fillable_mobile=True,
)

# Initialize the chat (with a system prompt and starting message)
chat = ui.Chat(
    id="chat",
    messages=[
        {"role": "system", "content": recipe_prompt},
        {
            "role": "assistant",
            "content": "Hello! I'm a recipe extractor. Please enter a URL to a recipe page. For example, <https://www.thechunkychef.com/epic-dry-rubbed-baked-chicken-wings/>",
        },
    ],
)

chat.ui(placeholder="Enter a recipe URL...")


# A function to transform user input
# Note that, if an exception occurs, the function will return a message to the user
# "short-circuiting" the conversation and asking the user to try again.
@chat.transform_user_input
async def try_scrape_page(input: str) -> str | None:
    try:
        return await scrape_page_with_url(input)
    except Exception:
        await chat.append_message(
            "I'm sorry, I couldn't extract content from that URL. Please try again. "
        )
        return None


@chat.on_user_submit
async def _():
    response = await llm.chat.completions.create(
        model="gpt-4o",
        messages=chat.messages(format="openai"),
        temperature=0,
        stream=True,
    )
    await chat.append_message_stream(response)



from __future__ import annotations

import re
from typing import Union, overload

__all__ = (
    "CssUnit",
    "as_css_unit",
    "as_css_padding",
    # "as_width_unit",
    # "isinstance_cssunit",
    # "as_grid_unit",
)

CssUnit = Union[
    int,
    float,
    str,
]
"""
Possible python types that can be converted into a CSS unit. Numeric values will be converted to pixels. Values equal to `0` will be converted to `"0"`. Strings will be passed through as-is.
"""


@overload
def as_css_unit(value: None) -> None: ...


@overload
def as_css_unit(value: CssUnit) -> str: ...


def as_css_unit(value: None | CssUnit) -> None | str:
    """
    Convert a value into a CSS unit.

    Parameters
    ----------
    value
        A value to convert into a CSS unit.

    Returns
    -------
    :
        If the `value` is `None`, then `None`. If the value is `0`, then `"0"`. If the `value` is numeric, then a formatted pixel value. Otherwise, the `value` as-is.

    Examples
    --------

    ```{python}
    from shiny.ui.css import as_css_unit

    as_css_unit(0)
    ```
    ```{python}
    as_css_unit(300)
    ```
    ```{python}
    as_css_unit("1em")
    ```
    """
    # TODO-future: Actually validate. Or don't validate, but then change
    # the function name to to_css_unit() or something.
    if isinstance(value, (float, int)):
        # Explicit check for 0 because floats may format to have many decimals.
        if value == 0:
            return "0"
        return "{:f}px".format(value)
    else:
        return value


@overload
def as_css_padding(padding: CssUnit | list[CssUnit]) -> str: ...


@overload
def as_css_padding(padding: None) -> None: ...


def as_css_padding(padding: CssUnit | list[CssUnit] | None) -> str | None:
    """
    Convert a CSS unit or list of CSS units into a CSS padding value.

    Parameters
    ----------
    padding
        A CSS unit or list of CSS units.

    Returns
    -------
    :
        A CSS padding value.

    Examples
    --------

    ```{python}
    from shiny.ui.css import as_css_padding

    as_css_padding([0, "1em"])
    ```
    """
    if padding is None:
        return None

    if not isinstance(padding, list):
        padding = [padding]

    return " ".join(as_css_unit(p) for p in padding)


# # It seems to be to use % over fr here since there is no gap on the grid
# def as_width_unit(x: str | float | int) -> str:
#     if isinstance(x, (int, float)):
#         return as_css_unit(x)

#     if isinstance(x, str) and x.endswith("%") and x.count("%") == 1:
#         x1_num = float(x[:-1])
#         x2_num = 100 - x1_num
#         return f"{x1_num}% {x2_num}%"

#     # TODO-bslib: validateCssUnit() should maybe support fr units?
#     # return(paste(x, collapse = " "))
#     return as_css_unit(x)


# Method for python legacy typing support. Can't just call `isinstance(x, CssUnit)` < python v3.10
def isinstance_cssunit(x: object) -> bool:
    return isinstance(x, (int, float, str))


@overload
def as_grid_unit(x: CssUnit) -> str: ...


@overload
def as_grid_unit(x: None) -> None: ...


def as_grid_unit(x: CssUnit | None) -> str | None:
    if x is None:
        return None

    if not isinstance(x, str):
        return as_css_unit(x)

    x_lower = x.lower()
    if x_lower in ("auto", "min-content", "max-content"):
        return x_lower
    if x.startswith("minmax("):
        return x
    if re.match(r"\d+\s*fr$", x):
        x = as_css_unit(x)
    return x



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, TypedDict, cast

import narwhals.stable.v1 as nw
import orjson

from ...session import Session, require_active_session
from ...types import Jsonifiable, JsonifiableDict
from ._html import as_cell_html, ui_must_be_processed
from ._types import (
    CellHtml,
    CellPatch,
    CellValue,
    ColsList,
    DataFrame,
    DataFrameT,
    DType,
    FrameDtype,
    FrameJson,
    IntoDataFrame,
    IntoDataFrameT,
    PandasCompatible,
    RowsList,
)

__all__ = (
    "as_data_frame",
    "data_frame_to_native",
    "apply_frame_patches",
    "serialize_dtype",
    "serialize_frame",
    "subset_frame",
)

if TYPE_CHECKING:
    import pandas as pd

########################################################################################
# Narwhals
#
# This module contains functions for working with data frames. It is a wrapper
# around the Narwhals library, which provides a unified interface for working with
# data frames (e.g. Pandas and Polars).
#
# The functions in this module are used to:
# * convert data frames to and from the Narwhals format,
# * apply patches to data frames,
# * serialize data frames to JSON,
# * subset data frames,
# * and get information about data frames (e.g. shape, column names).
#
# The functions in this module are used by the Shiny framework to work with data frames.
#
# For more information on the Narwhals library, see:
# * Intro https://narwhals-dev.github.io/narwhals/basics/dataframe/
# * `nw.DataFrame`: https://narwhals-dev.github.io/narwhals/api-reference/dataframe/
# * `nw.typing.IntoDataFrameT`: https://narwhals-dev.github.io/narwhals/api-reference/typing/#intodataframet
#
########################################################################################


# as_frame -----------------------------------------------------------------------------


def assert_data_is_not_none(data: IntoDataFrame) -> None:
    if data is None:  # pyright: ignore[reportUnnecessaryComparison]
        raise TypeError("`data` cannot be `None`")


def data_frame_to_native(data: DataFrame[IntoDataFrameT]) -> IntoDataFrameT:
    return nw.to_native(data)


def as_data_frame(
    data: IntoDataFrameT | DataFrame[IntoDataFrameT],
) -> DataFrame[IntoDataFrameT]:
    assert_data_is_not_none(data)

    if isinstance(data, DataFrame):
        return data  # pyright: ignore[reportUnknownVariableType]
    try:
        return nw.from_native(data, eager_only=True)
    except TypeError as e:
        try:
            compatible_data = compatible_to_pandas(data)
            ret: DataFrame[pd.DataFrame] = nw.from_native(
                compatible_data, eager_only=True
            )
            # Cast internal data as `IntoDataFrameT` type.
            # A warning has already been given to the user, so this is tolerable.
            return cast(DataFrame[IntoDataFrameT], ret)
        except TypeError:
            # Couldn't convert to pandas, so raise the original error
            raise e


def compatible_to_pandas(
    data: IntoDataFrame,
) -> pd.DataFrame:
    """
    Convert data to pandas, if possible.

    Should only call this method if Narwhals can not handle the data directly.
    """
    # Legacy support for non-Pandas/Polars data frames that were previously supported
    # and converted to pandas
    if isinstance(data, PandasCompatible):
        from ..._deprecated import warn_deprecated

        warn_deprecated(
            "A `.to_pandas()` was found on your data object and will be called. "
            "To remove this warning, please call `.to_pandas()` on your data "
            "and use the pandas result in your returned value. "
            "In the future, this will raise an error.",
            stacklevel=3,
        )
        return data.to_pandas()
        # pyright: ignore[reportReturnType]

    raise TypeError(f"Unsupported data type: {type(data)}")


# apply_frame_patches --------------------------------------------------------------------
def apply_frame_patches(
    nw_data: DataFrame[IntoDataFrameT],
    patches: List[CellPatch],
) -> DataFrame[IntoDataFrameT]:

    if len(patches) == 0:
        return nw_data

    # Apply the patches

    # Copy the data to make sure the original data is not modified in place.
    # If https://github.com/narwhals-dev/narwhals/issues/1154 is resolved, this
    # should be able to be removed.
    nw_data = nw_data.clone()

    # # https://discord.com/channels/1235257048170762310/1235257049626181656/1283415086722977895
    # # Using narwhals >= v1.7.0
    # @nw.narwhalify
    # def func(df):
    #     return df.with_columns(
    #         df['a'].scatter([0, 1], [999, 888]),
    #         df['b'].scatter([0, 1], [777, 555]),
    #     )

    # Group patches by column
    # This allows for a single column to be updated in a single operation (rather than multiple updates to the same column)
    #
    # In; patches: List[Dict[row_index: int, column_index: int, value: Any]]
    # Out; cell_patches_by_column: Dict[column_name: str, List[Dict[row_index: int, value: Any]]]
    #
    cell_patches_by_column: dict[str, ScatterValues] = {}
    for cell_patch in patches:
        column_name = nw_data.columns[cell_patch["column_index"]]
        if column_name not in cell_patches_by_column:
            cell_patches_by_column[column_name] = {
                "row_indexes": [],
                "values": [],
            }

        # Append the row index and value to the column information
        cell_patches_by_column[column_name]["row_indexes"].append(
            cell_patch["row_index"]
        )
        cell_patches_by_column[column_name]["values"].append(cell_patch["value"])

    # Upgrade the Scatter info to new column Series objects
    scatter_columns = [
        nw_data[column_name].scatter(
            scatter_values["row_indexes"], scatter_values["values"]
        )
        for column_name, scatter_values in cell_patches_by_column.items()
    ]
    # Apply patches to the nw data
    return nw_data.with_columns(*scatter_columns)


# serialize_dtype ----------------------------------------------------------------------
def serialize_dtype(col: nw.Series) -> FrameDtype:

    from ._html import series_contains_htmltoolslike

    dtype: DType = col.dtype

    if isinstance(dtype, nw.String):
        type_ = "string"

    elif dtype.is_numeric():
        type_ = "numeric"

    elif isinstance(dtype, (nw.Categorical, nw.Enum)):
        categories = col.cat.get_categories().to_list()
        return {"type": "categorical", "categories": categories}
    elif isinstance(dtype, nw.Boolean):
        type_ = "boolean"
    elif isinstance(dtype, nw.Date):
        type_ = "date"
    elif isinstance(dtype, nw.Datetime):
        type_ = "datetime"
    elif isinstance(dtype, nw.Duration):
        type_ = "duration"
    elif isinstance(dtype, nw.Object):
        type_ = "object"
        if series_contains_htmltoolslike(col):
            type_ = "html"
    elif isinstance(dtype, (nw.Struct, nw.List, nw.Array)):
        type_ = "object"
    else:
        type_ = "unknown"
        if series_contains_htmltoolslike(col):
            type_ = "html"

    return {"type": type_}


# serialize_frame ----------------------------------------------------------------------

RenderedDependency = dict[str, Jsonifiable]


def serialize_frame(into_data: IntoDataFrame) -> FrameJson:

    data = as_data_frame(into_data)

    type_hints = [serialize_dtype(data[col_name]) for col_name in data.columns]

    # TODO-future-barret; Swich serialization to "by column", rather than "by row"
    # * This would allow for a single column to be serialized in a single operation
    #   * This would allow for each column to capture the `"html"` type hint properly
    #     for object and unknown columns. Currently, there is no way to determine which
    #     cells are HTML-like during orjson serialization.
    # * Even better, would be to move `orjson` serialization to the websocket
    #   serialization.
    #   * No need to serialize to JSON, then unserialize, then serialize again when
    #     sending it to the client!
    #   * The only serialization would occur during "send to browser"
    #   * This approach would lose the ability to upgrade the column type hints to "html"
    data_rows = data.rows(named=False)

    session: Session | None = None
    html_deps: list[RenderedDependency] = []

    # Collect all html deps and dedupe them. Send the html separately from its deps
    # Otherwise, serialize as `str()`
    def default_orjson_serializer(val: Any) -> Jsonifiable:
        nonlocal session
        if ui_must_be_processed(val):
            if session is None:
                session = require_active_session(None)

            processed_ui = session._process_ui(val)
            deps = processed_ui["deps"]
            if len(deps) > 0:
                html_deps.extend(deps)
            # Remove the deps from the rendered html as we'll send them separately
            # Maintain expected structure so that the JS will attempt to add all
            #   dependencies for patched cells in the same manner
            processed_ui["deps"] = []
            cell_html_obj: CellHtml = as_cell_html(processed_ui)
            return cast(JsonifiableDict, cell_html_obj)

        # All other values are serialized as strings
        return str(val)

    data_val = orjson.loads(
        orjson.dumps(
            data_rows,
            default=default_orjson_serializer,
            # option=(orjson.OPT_NAIVE_UTC),
        )
    )

    deduped_html_deps = (
        _resolve_processed_dependencies(html_deps) if len(html_deps) > 1 else html_deps
    )

    return {
        "columns": data.columns,
        "data": data_val,
        "typeHints": type_hints,
        "htmlDeps": deduped_html_deps,
    }


# subset_frame -------------------------------------------------------------------------
def subset_frame(
    data: DataFrameT,
    *,
    rows: RowsList = None,
    cols: ColsList = None,
) -> DataFrameT:
    """Return a subsetted DataFrame, based on row positions and column names.

    Note that when None is passed, all rows or columns get included.
    """
    # Note that this type signature assumes column names are string objects.
    # This is always true in Polars, but not in Pandas (e.g. a column name could be an
    # int, or even a tuple of ints)

    # The nested if-else structure is used to navigate around narwhals' typing system and lack of `:` operator outside of `[`, `]`.

    if cols is None:
        if rows is None:
            return data
        else:
            return data[rows, :]
    else:
        # `cols` is not None
        col_names = [data.columns[col] if isinstance(col, int) else col for col in cols]

        # Return a DataFrame with no rows or columns
        # If there are no columns, there are no Series objects to support any rows
        if len(col_names) == 0:
            return data[[], []]

        if rows is None:
            return data[:, col_names]
        else:
            return data[rows, col_names]


class ScatterValues(TypedDict):
    row_indexes: list[int]
    values: list[CellValue]


# Direct copy of htmltools._core._resolve_dependencies
# Need new method as we need to access dep values via `dep[NAME]` rather than `dep.NAME`
def _resolve_processed_dependencies(
    deps: list[RenderedDependency],
) -> list[RenderedDependency]:
    map: dict[str, RenderedDependency] = {}
    for dep in deps:
        dep_name = str(dep["name"])
        if dep_name not in map:
            map[dep_name] = dep
        else:
            dep_version = str(dep["version"])
            cur_dep_version = str(map[dep_name]["version"])
            if dep_version > cur_dep_version:
                map[dep_name] = dep

    return list(map.values())



import pytest
from playwright.sync_api import Page, expect
from utils.deploy_utils import (
    local_deploys_app_url_fixture,
    reruns,
    reruns_delay,
    skip_if_not_chrome,
)

app_url = local_deploys_app_url_fixture("shiny-express-folium")


@skip_if_not_chrome
@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
def test_folium_map(page: Page, app_url: str) -> None:
    page.goto(app_url)

    expect(page.get_by_text("Static Map")).to_have_count(1)
    expect(page.get_by_text("Map inside of render express call")).to_have_count(1)
    # map inside the @render.express
    expect(
        page.frame_locator("iframe").nth(1).get_by_role("link", name="OpenStreetMap")
    ).to_have_count(1)
    # map outside of the @render.express at the top level
    expect(
        page.frame_locator("iframe")
        .nth(0)
        .get_by_role("link", name="U.S. Geological Survey")
    ).to_have_count(1)



from shiny import reactive, render
from shiny.express import input, ui

ui.input_action_button("same_id", "Action")
ui.input_action_button("same_id", "Action")


@render.text()
@reactive.event(input.same_id)
def counter():
    return f"{input.same_id()}"



from __future__ import annotations

import ast
import importlib.abc
import importlib.util
import sys
import types
from importlib.machinery import ModuleSpec
from pathlib import Path
from typing import Mapping, Sequence, cast

from htmltools import Tag, TagList

from .._app import App
from .._docstring import no_example
from .._typing_extensions import NotRequired, TypedDict
from .._utils import import_module_from_path
from ..session import Inputs, Outputs, Session, get_current_session, session_context
from ..types import MISSING, MISSING_TYPE
from ._is_express import find_magic_comment_mode
from ._recall_context import RecallContextManager
from ._stub_session import ExpressStubSession
from .expressify_decorator._func_displayhook import _expressify_decorator_function_def
from .expressify_decorator._node_transformers import (
    DisplayFuncsTransformer,
    expressify_decorator_func_name,
)

__all__ = (
    "app_opts",
    "wrap_express_app",
)

# Mapping from package name to file path of app. When running multiple concurrent apps
# (as is possible with shinylive), we need to give each one a unique package name.
package_filepath_map: dict[str, Path] = {}


@no_example()
def wrap_express_app(file: Path) -> App:
    """Wrap a Shiny Express mode app into a Shiny `App` object.

    This also creates a Python package for the app named something like
    `shiny_express_app_0`. This package is required for relative imports to work, as in
    `from . import utils`.

    Parameters
    ----------
    file
        The path to the file containing the Shiny express application.

    Returns
    -------
    :
        A :class:`shiny.App` object.
    """
    package_name = f"shiny_express_app_{len(package_filepath_map)}"
    package_filepath_map[package_name] = file

    # Importing the module triggers the ShinyExpressAppImportFinder and
    # ShinyExpressAppLoader.
    app_module = importlib.import_module(package_name)
    return app_module.app


# ======================================================================================
# Import hook to load Shiny Express app as a package
# ======================================================================================
class ShinyExpressAppImportFinder(importlib.abc.MetaPathFinder):
    def __init__(self):
        self.loaded_modules: dict[str, ModuleSpec] = {}

    def find_spec(
        self,
        fullname: str,
        path: Sequence[str] | None,
        target: types.ModuleType | None = None,
    ) -> ModuleSpec | None:
        if fullname in self.loaded_modules:
            return self.loaded_modules[fullname]

        if fullname in package_filepath_map:
            app_path = package_filepath_map[fullname]
            app_dir = str(app_path.parent)
            spec = importlib.util.spec_from_loader(
                fullname, ShinyExpressAppLoader(), origin=app_dir
            )
            if spec is None:
                return None

            spec.submodule_search_locations = [app_dir]
            self.loaded_modules[fullname] = spec
            return spec


class ShinyExpressAppLoader(importlib.abc.Loader):
    def create_module(self, spec: ModuleSpec):
        my_module = types.ModuleType(spec.name)
        return my_module

    def exec_module(self, module: types.ModuleType) -> None:
        module.app = create_express_app(  # pyright: ignore[reportAttributeAccessIssue]
            package_filepath_map[module.__name__], module.__name__
        )


sys.meta_path.insert(0, ShinyExpressAppImportFinder())
# ======================================================================================


# This is invoked from the ShinyExpressAppLoader.exec_module() method above. It creates
# the App object from the Shiny Express app file.
@no_example()
def create_express_app(file: Path, package_name: str) -> App:

    file = file.resolve()

    try:
        globals_file = file.parent / "globals.py"
        if globals_file.is_file():
            with session_context(None):
                import_module_from_path("globals", globals_file)

        stub_session = ExpressStubSession()
        with session_context(stub_session):
            # We tagify here, instead of waiting for the App object to do it when it wraps
            # the UI in a HTMLDocument and calls render() on it. This is because
            # AttributeErrors can be thrown during the tagification process, and we need to
            # catch them here and convert them to a different type of error, because uvicorn
            # specifically catches AttributeErrors and prints an error message that is
            # misleading for Shiny Express. https://github.com/posit-dev/py-shiny/issues/937
            app_ui = run_express(file, package_name).tagify()

    except AttributeError as e:
        raise RuntimeError(e) from e

    def express_server(input: Inputs, output: Outputs, session: Session):
        try:
            run_express(file, package_name)

        except Exception:
            import traceback

            # Starting in Python 3.10 this could be traceback.print_exception(e)
            traceback.print_exception(*sys.exc_info())
            raise

    app_opts: AppOpts = {}

    www_dir = file.parent / "www"
    if www_dir.is_dir():
        app_opts["static_assets"] = {"/": www_dir}

    app_opts = _merge_app_opts(app_opts, stub_session.app_opts)
    app_opts = _normalize_app_opts(app_opts, file.parent)

    app = App(
        app_ui,
        express_server,
        **app_opts,  # pyright: ignore[reportArgumentType]
    )

    return app


def run_express(file: Path, package_name: str | None = None) -> Tag | TagList:
    """
    Run the code in a Shiny Express app file and return the UI. This is to be run in
    both the UI-rendering phase and the server-rendering phase of a Shiny Express app.
    When used in the server-rendering phase, the returned UI should simply be ignored.

    Parameters
    ----------
    file
        The path to the file containing the Shiny Express application.
    package_name
        The name of the package for the app. This is generated by `wrap_express_app()`
        and should be something like "shiny_express_app_0". The purpose of this is to
        allow relative imports in the app code.
    """
    with open(file, encoding="utf-8") as f:
        content = f.read()

    tree = ast.parse(content, file)
    tree = DisplayFuncsTransformer().visit(tree)
    tree = ast.fix_missing_locations(tree)

    ui_result: Tag | TagList = TagList()

    def set_result(x: object):
        nonlocal ui_result
        ui_result = cast(Tag, x)

    prev_displayhook = sys.displayhook
    sys.displayhook = set_result

    try:
        reset_top_level_recall_context_manager()
        get_top_level_recall_context_manager().__enter__()

        file_path = str(file.resolve())

        var_context: dict[str, object] = {
            "__file__": file_path,
            "__name__": "app",
            "__package__": package_name,
            expressify_decorator_func_name: _expressify_decorator_function_def,
            "input": InputNotImportedShim(),
        }

        # Execute each top-level node in the AST
        for node in tree.body:
            if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                exec(
                    compile(ast.Module([node], type_ignores=[]), file_path, "exec"),
                    var_context,
                    var_context,
                )
            else:
                exec(
                    compile(ast.Interactive([node]), file_path, "single"),
                    var_context,
                    var_context,
                )

        # When we called the function to get the top level recall context manager, we didn't
        # store the result in a variable and re-use that variable here. That is intentional,
        # because during the evaluation of the app code,
        # replace_top_level_recall_context_manager() may have been called, which swaps
        # out the context manager, and it's the new one that we need to exit here.
        get_top_level_recall_context_manager().__exit__(None, None, None)

        # If we're running as an Express app but there's also a top-level item named app
        # which is a shiny.App object, the user probably made a mistake. (But if there's
        # a magic comment to force it into Express mode, don't raise, because that means
        # the user should know what they're doing.)
        if (
            "app" in var_context
            and isinstance(var_context["app"], App)
            and find_magic_comment_mode(content[:1000]) is None
        ):
            raise RuntimeError(
                "This looks like a Shiny Express app because it imports shiny.express, "
                "but it also looks like a Shiny Core app because it has a variable named "
                "`app` which is a shiny.App object. Remove either the shiny.express import, "
                "or the app=App()."
            )

        return ui_result

    except AttributeError as e:
        # Need to catch AttributeError and convert to a different type of error, because
        # uvicorn specifically catches AttributeErrors and prints an error message that
        # is helpful for normal ASGI apps, but misleading in the case of Shiny Express.
        raise RuntimeError(e) from e

    finally:
        sys.displayhook = prev_displayhook


_top_level_recall_context_manager: RecallContextManager[Tag] | None = None


def reset_top_level_recall_context_manager() -> None:
    from .ui._page import page_auto_cm

    global _top_level_recall_context_manager
    _top_level_recall_context_manager = page_auto_cm()


def get_top_level_recall_context_manager() -> RecallContextManager[Tag]:
    if _top_level_recall_context_manager is None:
        raise RuntimeError("No top-level recall context manager has been set.")

    return _top_level_recall_context_manager


class InputNotImportedShim:
    # This is a dummy class that is used to provide a helpful error message when the
    # user tries to access `input.x` but forgot to import `input`. If they do that, then
    # it would get the builtin `input` function, and print an unhelpful error message:
    #   RuntimeError: 'builtin_function_or_method' object has no attribute 'x'
    # This class provides an error message that is more helpful.
    def __getattr__(self, name: str):
        raise AttributeError(
            "Tried to access `input`, but it was not imported. "
            "Perhaps you need `from shiny.express import input`?"
        )


class AppOpts(TypedDict):
    static_assets: NotRequired[dict[str, Path]]
    debug: NotRequired[bool]


@no_example()
def app_opts(
    static_assets: str | Path | Mapping[str, str | Path] | MISSING_TYPE = MISSING,
    debug: bool | MISSING_TYPE = MISSING,
):
    """
    Set App-level options in Shiny Express

    This function sets application-level options for Shiny Express. These options are
    the same as those from the :class:`shiny.App` constructor.

    Parameters
    ----------
    static_assets
        Static files to be served by the app. If this is a string or Path object, it
        must be a directory, and it will be mounted at `/`. If this is a dictionary,
        each key is a mount point and each value is a file or directory to be served at
        that mount point. In Shiny Express, if there is a `www` subdirectory of the
        directory containing the app file, it will automatically be mounted at `/`, even
        without needing to set the option here.
    debug
        Whether to enable debug mode.
    """

    stub_session = get_current_session()

    if stub_session is None:
        # We can get here if a Shiny Core app, or if we're in the UI rendering phase of
        # a Quarto-Shiny dashboard.
        raise RuntimeError(
            "express.app_opts() can only be used in a standalone Shiny Express app."
        )

    # Store these options only if we're in the UI-rendering phase of Shiny Express.
    if not isinstance(stub_session, ExpressStubSession):
        return

    if not isinstance(static_assets, MISSING_TYPE):
        if isinstance(static_assets, (str, Path)):
            static_assets = {"/": Path(static_assets)}

        # Convert string values to Paths. (Need new var name to help type checker.)
        static_assets_paths = {k: Path(v) for k, v in static_assets.items()}

        stub_session.app_opts["static_assets"] = static_assets_paths

    if not isinstance(debug, MISSING_TYPE):
        stub_session.app_opts["debug"] = debug


def _merge_app_opts(app_opts: AppOpts, app_opts_new: AppOpts) -> AppOpts:
    """
    Merge a set of app options into an existing set of app options. The values from
    `app_opts_new` take precedence. This will alter the original app_opts and return it.
    """

    # We can't just do a `app_opts.update(app_opts_new)` because we need to handle the
    # case where app_opts["static_assets"] and app_opts_new["static_assets"] are
    # dictionaries, and we need to merge those dictionaries.
    if "static_assets" in app_opts and "static_assets" in app_opts_new:
        app_opts["static_assets"].update(app_opts_new["static_assets"])
    elif "static_assets" in app_opts_new:
        app_opts["static_assets"] = app_opts_new["static_assets"].copy()

    if "debug" in app_opts_new:
        app_opts["debug"] = app_opts_new["debug"]

    return app_opts


def _normalize_app_opts(app_opts: AppOpts, parent_dir: Path) -> AppOpts:
    """
    Normalize the app options, ensuring that all paths in static_assets are absolute.
    Modifies the original in place.
    """
    if "static_assets" in app_opts:
        for mount_point, path in app_opts["static_assets"].items():
            if not path.is_absolute():
                path = parent_dir / path
            app_opts["static_assets"][mount_point] = path

    return app_opts



from __future__ import annotations

from typing import Literal, TypeVar, overload

from ._docstring import add_example
from .types import (
    SilentCancelOutputException,
    SilentException,
    SilentOperationInProgressException,
)

T = TypeVar("T")


@overload
def req(*, cancel_output: bool | Literal["progress"] = False) -> None: ...


@overload
def req(*args: T, cancel_output: bool | Literal["progress"] = False) -> T: ...


@add_example()
def req(*args: T, cancel_output: bool | Literal["progress"] = False) -> T | None:
    """
    Throw a silent exception for falsy values.

    This is a convenient shorthand for throwing :class:`~shiny.types.SilentException` /
    :class:`~shiny.types.SilentCancelOutputException` if any of the arguments are falsy.

    The term "falsy" generally indicates that a value is considered `False` when
    encountered in a logical context. We use the term a little loosely here; our usage
    tries to match the intuitive notions of "Is this value missing or available?", or
    "Has the user provided an answer?", or in the case of action buttons, "Has the
    button been clicked?". So `False`, `None`, `0`, and `""` would be examples of Falsy
    values.

    Parameters
    ----------
    *args
        Any number of arguments to check.
    cancel_output
        If ``True``, throw :class:`~shiny.types.SilentCancelOutputException` instead of
        :class:`~shiny.types.SilentException`.

    Returns
    -------
    :
        The first argument. If no arguments are provided, returns ``None``.
    """
    if len(args) == 0:
        return None

    for arg in args:
        if not arg:
            if cancel_output is True:
                raise SilentCancelOutputException()
            elif cancel_output == "progress":
                raise SilentOperationInProgressException()
            else:
                raise SilentException()

    return args[0]



import folium

from shiny import render
from shiny.express import input, ui

locations_coords = {
    "San Francisco": (37.79554, -122.39348),
    "Los Angeles": (34.05026, -118.25768),
    "New York": (40.71222, -74.00490),
}
ui.page_opts(full_width=False)

with ui.card(id="card"):
    "Static Map"
    folium.Map(  # pyright: ignore[reportUnknownMemberType,reportGeneralTypeIssues]
        location=locations_coords["San Francisco"], tiles="USGS.USTopo", zoom_start=12
    )
    ui.input_radio_buttons(
        "location", "Location", ["San Francisco", "New York", "Los Angeles"]
    )

    @render.express
    def folium_map():
        "Map inside of render express call"
        folium.Map(  # pyright: ignore[reportUnknownMemberType,reportGeneralTypeIssues]
            location=locations_coords[input.location()],
            tiles="cartodb positron",
            zoom_start=12,
        )
        input.location()



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.playwright.expect import expect_to_have_style
from shiny.run import ShinyAppProc


def test_accordion(page: Page, local_app: ShinyAppProc, is_webkit: bool) -> None:
    page.goto(local_app.url)

    text = controller.OutputCode(page, "text")
    tab = controller.NavsetTab(page, "tab")

    test_text_area = controller.InputTextArea(page, "test_text_area")
    test_text_area_w_rows = controller.InputTextArea(page, "test_text_area2")

    text.expect_value("Loaded")

    # Make sure the `rows` is respected
    test_text_area_w_rows.expect_rows("4")
    # Make sure the placeholder row value of `1` is set
    test_text_area.expect_rows("1")

    tab.set("Text Area")

    test_text_area.expect_autoresize(True)
    test_text_area.expect_value("a\nb\nc\nd\ne")

    if is_webkit:
        # Skip the rest of the test for webkit.
        # Heights are not consistent with chrome and firefox
        return
    expect_to_have_style(test_text_area.loc, "height", "125px")
    expect_to_have_style(test_text_area_w_rows.loc, "height", "125px")

    # Make sure the `rows` is consistent
    test_text_area.expect_rows("1")
    test_text_area_w_rows.expect_rows("4")

    # Reset the text area to a single row and make sure the area shrink to appropriate size
    test_text_area.set("single row")
    test_text_area_w_rows.set("single row")

    # 1 row
    expect_to_have_style(test_text_area.loc, "height", "35px")
    # 4 rows
    expect_to_have_style(test_text_area_w_rows.loc, "height", "102px")



from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_dataframe_methods(page: Page, local_app: ShinyAppProc) -> None:

    page.goto(local_app.url)

    df_html = controller.OutputDataFrame(page, "df_html")

    df_html.expect_nrow(2)
    df_html.expect_ncol(12)

    df_html.expect_column_labels(
        [
            "num",
            "chr",
            "cat",
            "bool",
            "date",
            "datetime",
            "duration",
            "html",
            "html_str",
            "struct",
            "arr",
            "object",
        ]
    )

    df_html.expect_cell("1", row=0, col=0)
    df_html.expect_cell("2", row=1, col=0)

    df_html.expect_cell("a", row=0, col=1)
    df_html.expect_cell("b", row=1, col=1)

    df_html.expect_cell("c", row=0, col=2)
    df_html.expect_cell("d", row=1, col=2)

    df_html.expect_cell("true", row=0, col=3)
    df_html.expect_cell("false", row=1, col=3)

    df_html.expect_cell("2000-01-02", row=0, col=4)
    df_html.expect_cell("2000-01-02", row=1, col=4)

    df_html.expect_cell("2000-01-02 00:00:00", row=0, col=5)
    df_html.expect_cell("2000-01-02 00:00:00", row=1, col=5)

    df_html.expect_cell("7 days 00:00:00", row=0, col=6)
    df_html.expect_cell("7 days 00:00:00", row=1, col=6)

    df_html.expect_cell("bolded content", row=0, col=7)
    df_html.expect_cell("bolded content", row=1, col=7)
    # Verify that is cell has a **strong** tag and isn't a string
    expect(df_html.cell_locator(row=0, col=7).locator("strong")).to_have_text(
        "bolded content"
    )

    df_html.expect_cell("bolded string", row=0, col=8)
    df_html.expect_cell("bolded string", row=1, col=8)
    # Verify that is cell has a **strong** tag and isn't a string
    expect(df_html.cell_locator(row=0, col=8).locator("strong")).to_have_text(
        "bolded string"
    )

    df_html.expect_cell('{"x":1}', row=0, col=9)
    df_html.expect_cell('{"x":2}', row=1, col=9)

    df_html.expect_cell("[1,2]", row=0, col=10)
    df_html.expect_cell("[3,4]", row=1, col=10)

    df_html.expect_cell("<C object>", row=0, col=11)
    df_html.expect_cell('{"y":2}', row=1, col=11)



from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Literal, Optional, TypeVar


def find_api_examples_dir(start_dir: str) -> Optional[str]:
    current_dir = os.path.abspath(start_dir)
    while True:
        api_examples_dir = os.path.join(current_dir, "api-examples")
        if os.path.isdir(api_examples_dir):
            return api_examples_dir
        root_files = ["setup.cfg", "pyproject.toml"]
        dir_files = os.listdir(current_dir)
        if any(rf in dir_files for rf in root_files):
            break  # Reached the package root directory
        if current_dir == os.path.dirname(current_dir):
            break  # Reached the global root directory
        current_dir = os.path.dirname(current_dir)
    return None


FuncType = Callable[..., Any]
F = TypeVar("F", bound=FuncType)


def no_example(mode: Optional[Literal["express", "core"]] = None) -> Callable[[F], F]:
    """
    Prevent ``@add_example()`` from throwing an error about missing examples.

    Parameters
    ----------
    mode:
        If ``"express"``, ``@add_example()`` will not throw an error if the current
        mode is Express. If ``"core"``, ``@add_example()`` will not throw an error if
        the current mode is Core. If ``None``, ``@add_example()`` will not throw an
        error in either mode.
    """

    def decorator(func: F) -> F:
        current = getattr(func, "__no_example", [])
        if mode is None:
            current.extend(["express", "core"])
        else:
            current.append(mode)
        setattr(func, "__no_example", current)  # noqa: B010
        return func

    return decorator


# This class is used to mark docstrings when @add_example() is used, so that an error
# will be thrown if @doc_format() is used afterward. This is to avoid an error when
# the example contains curly braces -- the @doc_format() decorator will try to evaluate
# the code in {}.
class DocStringWithExample(str): ...


class ExampleWriter:
    def write_example(self, app_files: list[str]) -> str:
        app_file = app_files[0]
        with open(app_file, encoding="utf-8") as f:
            code = f.read()

        return f"```.python\n{code.strip()}\n```\n"


example_writer = ExampleWriter()


def add_example(
    app_file: Optional[str] = None,
    ex_dir: Optional[str] = None,
) -> Callable[[F], F]:
    """
    Add an example to the docstring of a function, method, or class.

    This decorator must, at the moment, be used on a function, method, or class whose
    ``__name__`` matches the name of directory under a ``api-examples/`` directory in
    the current or any parent directory.

    * Examples for the ``shiny`` package are in ``shiny/api-examples/``. We also place
      Express examples in this directory adjacent to their Core counterparts.
    * Examples for the ``shiny.experimental`` subpackage are in
      ``shiny/experimental/api-examples/``.

    Functions that can be used in Express or Core and whose canonical implementation is
    in the ``shiny`` package should have examples in ``shiny/api-examples``. In this
    case, the express variant should include an ``-express`` suffix and the core
    variation can be named with a ``-core`` suffix or ``app.py``.

    Parameters
    ----------
    app_file:
        The primary app file to use for the example. This allows you to have multiple
        example files for a single function or to use a different file name than
        ``app.py``. Support files _cannot_ be named ``app.py`` or start with ``app-``,
        as these files will never be included in the example.
    ex_dir:
        The directory containing the example. If not specified, ``add_example()`` will
        find a directory named after the current function in the first ``api-examples/``
        directory it finds in the current directory or its parent directories.
    """

    def _(func: F) -> F:
        # To avoid a performance hit on `import shiny`, we only add examples to the
        # docstrings if this env variable is set (as it is in `make quartodoc`).
        if os.getenv("SHINY_ADD_EXAMPLES") != "true":
            if func.__doc__ is not None:
                func.__doc__ = DocStringWithExample(func.__doc__)
            return func

        current_mode = os.getenv("SHINY_MODE", "core")
        if current_mode in getattr(func, "__no_example", []):
            return func

        func_dir = get_decorated_source_directory(func)
        fn_name = func.__name__

        if ex_dir is None:
            ex_dir_found = find_api_examples_dir(func_dir)

            if ex_dir_found is None:
                raise FileNotFoundError(
                    f"No example directory found for {fn_name} in {func_dir} or its parent directories."
                )
            example_dir = os.path.join(ex_dir_found, fn_name)
        else:
            example_dir = os.path.abspath(os.path.join(func_dir, ex_dir))

            if not os.path.exists(example_dir):
                raise FileNotFoundError(
                    f"Example directory '{example_dir}' does not exist for {fn_name}."
                )

        app_file_name = app_file or "app.py"
        try:
            example_file = app_choose_core_or_express(
                os.path.join(example_dir, app_file_name),
                mode="express" if "shiny/express/" in func_dir else None,
            )
        except ExampleNotFoundException as e:
            file = "shiny/" + func_dir.split("shiny/")[1]
            if "__code__" in dir(func):
                print(
                    f"::warning file={file},line={func.__code__.co_firstlineno}::{fn_name} - {e}"
                )
            else:
                print(f"::warning file={file}::{fn_name} - {e}")

            return func

        other_files: list[str] = []
        for abs_f in Path(example_dir).glob("**/*"):
            rel_f = abs_f.relative_to(example_dir)
            path_parts = Path(rel_f).parts
            f = os.path.basename(rel_f)
            is_support_file = (
                os.path.isfile(abs_f)
                and f != app_file_name
                and f != "app.py"
                and f != ".DS_Store"
                and "venv" not in path_parts
                and ".venv" not in path_parts
                and not (f.startswith("app-") and f.endswith(".py"))
                and not str(rel_f).startswith("__")
            )
            if is_support_file:
                other_files.append(str(abs_f))

        if func.__doc__ is None:
            func.__doc__ = ""

        example = example_writer.write_example([example_file, *other_files])
        example_lines = example.split("\n")

        # How many leading spaces does the docstring start with?
        doc = func.__doc__.replace("\n", "")
        indent = " " * (len(doc) - len(doc.lstrip()))
        nl_indent = "\n" + indent

        # Add example header if not already present
        # WARNING: All `add_example()` calls must be coalesced.
        # Note that we're using numpydoc-style headers here, quartodoc will handle
        # converting them to markdown headers.
        if isinstance(func.__doc__, DocStringWithExample):
            ex_header = "Examples" + nl_indent + "--------"
            before, after = func.__doc__.split(ex_header, 1)
            func.__doc__ = before + ex_header
        else:
            func.__doc__ += nl_indent + "Examples"
            func.__doc__ += nl_indent + "--------"
            after = None

        # Insert the example under the Examples heading
        func.__doc__ += nl_indent * 2
        func.__doc__ += nl_indent.join(example_lines)
        if after is not None:
            func.__doc__ += after

        func.__doc__ = DocStringWithExample(func.__doc__)
        return func

    return _


def is_express_app(app_path: str) -> bool:
    # We can't use .shiny.express._is_express.is_express_app() here because that would
    # create a circular import.
    if not os.path.exists(app_path):
        return False

    with open(app_path, encoding="utf-8") as f:
        for line in f:
            if "from shiny.express" in line:
                return True
            elif "import shiny.express" in line:
                return True
    return False


class ExampleNotFoundException(FileNotFoundError):
    def __init__(
        self,
        file_names: list[str] | str,
        dir: str,
        type: Optional[Literal["core", "express"]] = None,
    ) -> None:
        self.type = type or os.environ.get("SHINY_MODE", "core")
        self.file_names = [file_names] if isinstance(file_names, str) else file_names
        self.dir = dir

    def __str__(self):
        if self.type in ("core", "express"):
            # Capitalize first letter
            type = "a Shiny Express" if self.type == "express" else "a Shiny Core"
        else:
            type = "an"

        return (
            f"Could not find {type} example file named "
            + f"{' or '.join(self.file_names)} in {self.dir}."
        )


class ExpressExampleNotFoundException(ExampleNotFoundException):
    def __init__(
        self,
        file_names: list[str] | str,
        dir: str,
    ) -> None:
        super().__init__(file_names, dir, "express")


def app_choose_core_or_express(
    app_path: Optional[str] = None,
    mode: Optional[Literal["express", "core"]] = None,
) -> str:
    app_path = app_path or "app.py"

    if mode is None:
        mode_env = os.environ.get("SHINY_MODE", "core")
        mode = "express" if mode_env == "express" else "core"

    if mode == "express":
        if is_express_app(app_path):
            return app_path

        app_path = app_path.replace("-core.py", ".py")

        path, ext = os.path.splitext(app_path)
        app_path_express = f"{path}-express{ext}"

        if not is_express_app(app_path_express):
            raise ExpressExampleNotFoundException(
                [os.path.basename(app_path), os.path.basename(app_path_express)],
                os.path.dirname(app_path),
            )

        return app_path_express

    if os.path.basename(app_path) == "app.py" and not os.path.exists(app_path):
        app_path = app_path.replace("app.py", "app-core.py")

    if not os.path.exists(app_path):
        raise ExampleNotFoundException(
            os.path.basename(app_path),
            os.path.dirname(app_path),
        )

    return app_path


def get_decorated_source_directory(func: FuncType) -> str:
    if hasattr(func, "__module__"):
        path = os.path.abspath(str(sys.modules[func.__module__].__file__))
    else:
        path = os.path.abspath(func.__code__.co_filename)

    return os.path.dirname(path)


def doc_format(**kwargs: str) -> Callable[[F], F]:
    def _(func: F) -> F:
        if isinstance(func.__doc__, DocStringWithExample):
            raise ValueError(
                f"@doc_format() must be applied before @add_example() for {func.__name__}."
            )
        if func.__doc__:
            func.__doc__ = func.__doc__.format(**kwargs)
        return func

    return _


if not TYPE_CHECKING and os.environ.get("IN_QUARTODOC") == "true":
    # When running in quartodoc, we use shinylive to embed the examples in the docs.
    # This part is hidden from the typechecker because shinylive is not a direct
    # dependency of shiny and we only need this section when building the docs.
    try:
        from shinylive import ShinyliveApp
    except ImportError:
        raise RuntimeError(
            "Please install the latest version of shinylive to build the docs."
        )
    except ModuleNotFoundError:
        raise RuntimeError("Please install shinylive to build the docs.")

    class ShinyliveExampleWriter(ExampleWriter):
        def write_example(self, app_files: list[str]) -> str:
            app_file = app_files.pop(0)
            app = ShinyliveApp.from_local(app_file, app_files, language="py")

            return app.to_chunk(layout="vertical", viewer_height=400)

    example_writer = ShinyliveExampleWriter()



import matplotlib.pyplot as plt
import numpy as np

from shiny.express import input, render, ui

ui.input_slider("n", "input_slider()", min=10, max=100, value=50, step=5, animate=True)


@render.plot
def p():
    np.random.seed(19680801)
    x_rand = 100 + 15 * np.random.randn(437)
    fig, ax = plt.subplots()
    ax.hist(x_rand, int(input.n()), density=True)
    return fig



import re
from importlib import util
from pathlib import Path

from prompt_toolkit.document import Document
from questionary import ValidationError, Validator


def is_existing_module(name: str) -> bool:
    """
    Check if a module name can be imported, which indicates that it is either
    a standard module name, or the name of an installed module.
    In either case the new module would probably cause a name conflict.
    """
    try:
        spec = util.find_spec(name)
        if spec is not None:
            return True
        else:
            return False
    except ImportError:
        return False


def is_pep508_identifier(name: str):
    """
    Checks if a package name is a PEP 508 identifier.
    """
    # Regex from https://peps.python.org/pep-0508/#names
    pattern = re.compile(r"^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$", re.IGNORECASE)
    return bool(pattern.match(name))


class ComponentNameValidator(Validator):
    def validate(self, document: Document):
        """
        Validate that the name is all lowercase and dash or space-delimited
        """
        name = document.text

        # Dont accept empty names
        if len(name) == 0:
            raise ValidationError(
                message="Name needed for component",
                cursor_position=len(name),
            )

        # Check for underscores
        if "_" in name:
            raise ValidationError(
                message="Use dashes `-` instead of underscores `_`",
                cursor_position=len(name),
            )

        # Check for uppercase letters
        if name != name.lower():
            raise ValidationError(
                message="Name must be all lowercase",
                cursor_position=len(name),
            )

        # Check for quotations

        if (
            name.startswith('"')
            or name.endswith('"')
            or name.startswith("'")
            or name.endswith("'")
        ):
            raise ValidationError(
                message="The name should be unquoted.",
                cursor_position=len(name),
            )

        # Pypi only allows names shorter than 214 characters
        if len(name) > 214:
            raise ValidationError(message="Name can't exceed 214 characters")

        if not is_pep508_identifier(name):
            raise ValidationError(
                message="Name must be a pep508 identifier: https://peps.python.org/pep-0508/#names",
                cursor_position=len(name),
            )

        # Using the name of an existing package causes an import error

        if is_existing_module(name):
            raise ValidationError(
                message="Package already installed in your current environment.",
                cursor_position=len(name),
            )


def update_component_name_in_template(template_dir: Path, new_component_name: str):
    """
    Function to update the name of a custom component in the template files
    Needs to find and replace the following:
    - `custom_component` -> `new_component_name`  # python module/pkg name
    - `custom-component` -> `new-component-name`  # component tag name
    - `CustomComponentEl` -> `NewComponentNameEl` # component class name
    Also needs to change some file/directory names
    - `customComponent/` -> `newComponentName/`         # python pkg name
    - `custom_component.py` -> `new_component_name.py`  # python module name
    """

    # Previous names
    old_underscore_name = "custom_component"
    old_dash_name = "custom-component"
    old_capital_case_name = "CustomComponent"

    # Create a series of names corresponding to the various formats needed
    # The name is currently provided in space-demlimited or dash-delimited format
    # e.g. "new-component-name" -> ["new", "component", "name"]
    #      "new component name" -> ["new", "component", "name"]

    # First normalize the name to be space-delimited
    # e.g. "new-component-name" -> "new component name"
    #      "new component name" -> "new component name"
    # and then split into an array of words
    name_parts = new_component_name.replace("-", " ").split(" ")

    # Now create the various formats needed
    underscore_name = "_".join(name_parts)
    dash_name = "-".join(name_parts)
    capital_case_name = "".join([part.capitalize() for part in name_parts])

    # Rename the directory containing the python code: aka ./custom_component -> ./new_component_name
    python_pkg_dir = template_dir / old_underscore_name
    python_pkg_dir.rename(template_dir / underscore_name)

    # Next rename the python module: aka ./new_component_name/custom_component.py -> ./new_component_name/new_component_name.py
    python_module_path = template_dir / underscore_name / f"{old_underscore_name}.py"
    python_module_path.rename(template_dir / underscore_name / f"{underscore_name}.py")

    def update_names_in_files(dir: Path):
        # Now do a find-and-replace for the various name types across all the files in the
        # template directory
        for item in dir.iterdir():
            if item.is_file():
                # Only do this for files
                with open(item, "r", encoding="utf-8") as f:
                    file_contents = f.read()
                # First, "custom_component" -> "new_component_name"
                file_contents = file_contents.replace(
                    old_underscore_name, underscore_name
                )
                # Next, "custom-component" -> "new-component-name"
                file_contents = file_contents.replace(old_dash_name, dash_name)
                # Next, "CustomComponentEl" -> "NewComponentNameEl"
                file_contents = file_contents.replace(
                    old_capital_case_name, capital_case_name
                )

                with open(item, "w", encoding="utf-8") as f:
                    f.write(file_contents)

    # Loop over dirs_to_update and run the update function on them
    for dir_to_update in ["", "srcts", underscore_name, "example-app"]:
        update_names_in_files(template_dir / dir_to_update)



from __future__ import annotations

from typing import TypeVar

from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page

from shiny.run import ShinyAppProc

T = TypeVar("T")

app = create_doc_example_core_fixture("layout_columns")


def not_null(x: T | None) -> T:
    assert x is not None
    return x


def test_layout_columns(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    col1 = page.locator("#loss_over_time")
    col2 = page.locator("#accuracy_over_time")
    col3 = page.locator("#feature_importance")

    width1, width2, width3 = [
        not_null(x.bounding_box()).get("width") for x in [col1, col2, col3]
    ]

    assert width1 > 400
    assert width1 < width2
    assert width1 + width2 < width3



import asyncio

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = ui.page_fluid(
    ui.tags.p(
        """
      The first time you click the button, you should see a 1 appear below the button,
      as well as 2 messages in the python console (all reporting 1 click). After
      clicking once, clicking again should increment the number below the button and
      print the number of clicks in the console twice.
      """
    ),
    ui.navset_card_tab(
        ui.nav_panel(
            "Sync",
            ui.input_action_button("btn", "Click me"),
            ui.output_ui("btn_value"),
        ),
        ui.nav_panel(
            "Async",
            ui.input_action_button("btn_async", "Click me"),
            ui.output_ui("btn_async_value"),
        ),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.btn)
    def _():
        print("@effect() event: ", str(input.btn()))

    @reactive.calc
    @reactive.event(input.btn)
    def btn() -> int:
        return input.btn()

    @reactive.effect
    def _():
        print("@calc() event:   ", str(btn()))

    @render.ui
    @reactive.event(input.btn)
    def btn_value():
        return str(input.btn())

    # -----------------------------------------------------------------------------
    # Async
    # -----------------------------------------------------------------------------
    @reactive.effect
    @reactive.event(input.btn_async)
    async def _():
        await asyncio.sleep(0)
        print("async @effect() event: ", str(input.btn_async()))

    @reactive.calc
    @reactive.event(input.btn_async)
    async def btn_async_r() -> int:
        await asyncio.sleep(0)
        return input.btn_async()

    @reactive.effect
    async def _():
        val = await btn_async_r()
        print("async @calc() event:   ", str(val))

    @render.ui
    @reactive.event(btn_async_r)
    async def btn_async_value():
        val = await btn_async_r()
        print("== " + str(val))
        return str(val)


app = App(app_ui, server)



from __future__ import annotations

from typing import Literal, Optional

from shiny.render.renderer import Renderer, ValueFn
from shiny.ui import output_text_verbatim


class render_capitalize(Renderer[str]):
    # The documentation for the class will be displayed when the user hovers over the
    # decorator when **no** parenthesis are used. Ex: `@render_capitalize`
    # If no documentation is supplied to the `__init__()` method, then this
    # documentation will be displayed when parenthesis are used on the decorator.
    """
    Render capitalize class documentation goes here.
    """

    to_case: Literal["upper", "lower", "ignore"]
    """
    The case to render the value in.
    """
    placeholder: bool
    """
    Whether to render a placeholder value. (Defaults to `True`)
    """

    def auto_output_ui(self):
        """
        Express UI for the renderer
        """
        return output_text_verbatim(self.output_id, placeholder=True)

    def __init__(
        self,
        _fn: Optional[ValueFn[str]] = None,
        *,
        to_case: Literal["upper", "lower", "ignore"] = "upper",
        placeholder: bool = True,
    ) -> None:
        # If a different set of documentation is supplied to the `__init__` method,
        # then this documentation will be displayed when parenthesis are used on the decorator.
        # Ex: `@render_capitalize()`
        """
        Render capitalize documentation goes here.

        It is a good idea to talk about parameters here!

        Parameters
        ----------
        to_case
            The case to render the value. (`"upper"`)

            Options:
            - `"upper"`: Render the value in upper case.
            - `"lower"`: Render the value in lower case.
            - `"ignore"`: Do not alter the case of the value.

        placeholder
            Whether to render a placeholder value. (`True`)
        """
        # Do not pass params
        super().__init__(_fn)
        self.to_case = to_case
        self.placeholder = placeholder

    async def render(self) -> str | None:
        value = await self.fn()
        if value is None:
            # If `None` is returned, then do not render anything.
            return None

        ret = str(value)
        if self.to_case == "upper":
            return ret.upper()
        if self.to_case == "lower":
            return ret.lower()
        if self.to_case == "ignore":
            return ret
        raise ValueError(f"Invalid value for `to_case`: {self.to_case}")


class render_upper(Renderer[str]):
    """
    Minimal capitalize string transformation renderer.

    No parameters are supplied to this renderer. This allows us to skip the `__init__()`
    method and `__init__()` documentation. If you hover over this decorator with and
    without parenthesis, you will see this documentation in both situations.

    Note: This renderer is equivalent to `render_capitalize(to="upper")`.
    """

    def auto_output_ui(self):
        """
        Express UI for the renderer
        """
        return output_text_verbatim(self.output_id, placeholder=True)

    async def transform(self, value: str) -> str:
        """
        Transform the value to upper case.

        This method is shorthand for the default `render()` method. It is useful to
        transform non-`None` values. (Any `None` value returned by the app author will
        be forwarded to the browser.)

        Parameters
        ----------
        value
            The a non-`None` value to transform.

        Returns
        -------
        str
            The transformed value. (Must be a subset of `Jsonifiable`.)
        """

        return str(value).upper()



from shiny.express import ui

ui.tags.style(
    """
    #shell div{
    background-color: #00000022}
"""
)
with ui.div(id="shell"):
    with ui.layout_columns(col_widths=[8, 4]):
        "R1C1R1"
        with ui.layout_columns(col_widths=[8, 4]):
            with ui.div():
                ui.div("R1C1R2-R1C1R1")

                ui.div("R1C1R2-R1C1R2")

            "R1C1R2-R1C2"

        "R1C2"

with ui.layout_columns(col_widths=[6, 6]):
    # check height is below 300px - bounding box
    with ui.navset_card_tab(id="express_navset_card_tab"):
        with ui.nav_panel(title="Two"):
            ...

    with ui.navset_tab(id="express_navset_tab"):
        for fn_txt, fn in [
            ("pre", ui.pre),
            ("div", ui.div),
            ("span", ui.span),
        ]:
            with ui.nav_panel(title=fn_txt):
                for i in range(3):
                    with fn():
                        ui.HTML(f"{fn_txt} {i}")



from __future__ import annotations

from pathlib import Path
from typing import List, Union

from htmltools import HTML, HTMLDependency, Tagifiable, TagList, head_content
from htmltools.tags import link

from .._versions import bootstrap as bootstrap_version
from ..html_dependencies import jquery_deps, shiny_deps
from ._html_deps_shinyverse import (
    components_dependencies as bslib_component_dependencies,
)
from ._include_helpers import check_path, include_css
from ._theme import Theme

"""
HTML dependencies for external dependencies Bootstrap, ionrangeslider, datepicker, selectize, and jQuery UI.

For...
* External dependencies (e.g. jQuery, Bootstrap), see `shiny.ui._html_deps_external`
* Internal dependencies (e.g. dataframe, autoresize), see `shiny.ui._html_deps_py_shiny`
* shinyverse dependencies (e.g. bslib, htmltools), see `shiny.ui._html_deps_shinyverse`
"""

ThemeProvider = Union[Tagifiable, HTMLDependency, List[HTMLDependency]]


def shiny_page_theme_deps(theme: str | Path | Theme | ThemeProvider | None) -> TagList:
    deps_bootstrap = bootstrap_deps(include_css=theme is None)

    if theme is None:
        deps_theme = None
    elif isinstance(theme, Theme):
        deps_theme = theme._html_dependencies()
    elif isinstance(theme, str) and theme.startswith(("http", "//")):
        deps_theme = head_content(link(rel="stylesheet", href=theme, type="text/css"))
    elif isinstance(theme, (str, Path)):
        check_path(theme)
        deps_theme = head_content(include_css(theme))
    elif isinstance(theme, Tagifiable) or isinstance(theme, HTMLDependency):
        deps_theme = theme
    elif isinstance(theme, list) and all(
        [isinstance(dep, HTMLDependency) for dep in theme]
    ):
        deps_theme = theme
    else:
        raise ValueError(
            "Invalid `theme`. "
            + "Expected a URL or path to a full Bootstrap CSS file, "
            + "or a theme provider, "
            + f"but received `theme` with type {type(theme)}."
        )

    return TagList(
        deps_bootstrap,
        deps_theme,
        # These next dependencies have their CSS bundled in the main Bootstrap CSS
        # but are also included at the component level. This page-level dependency will
        # win in Shiny apps, but the component-level dependencies will win in other
        # contexts where we don't have a page function.
        #
        # Note: this approach is only intended for legacy bundled components. If you're
        # contemplating following this approach for a new component, please reconsider
        # and explore styling the component with CSS variables instead.
        shiny_deps(include_css=False),
        bslib_component_dependencies(include_css=False),
        ionrangeslider_deps(include_css=False),
        datepicker_deps(include_css=False),
        selectize_deps(include_css=False),
    )


def bootstrap_deps(include_css: bool = True) -> list[HTMLDependency]:
    dep = HTMLDependency(
        name="bootstrap",
        version=bootstrap_version,
        source={"package": "shiny", "subdir": "www/shared/bootstrap/"},
        script={"src": "bootstrap.bundle.min.js"},
        stylesheet={"href": "bootstrap.min.css"} if include_css else None,
        meta={"name": "viewport", "content": "width=device-width, initial-scale=1"},
        all_files=True,
    )
    deps = [jquery_deps(), dep]
    return deps


def ionrangeslider_deps(include_css: bool = True) -> list[HTMLDependency]:
    return [
        HTMLDependency(
            name="ionrangeslider",
            version="2.3.1",
            source={"package": "shiny", "subdir": "www/shared/ionrangeslider/"},
            script={"src": "js/ion.rangeSlider.min.js"},
            # This CSS is rendered against default Bootstrap
            stylesheet={"href": "css/ion.rangeSlider.css"} if include_css else None,
        ),
        HTMLDependency(
            name="strftime",
            version="0.9.2",
            source={"package": "shiny", "subdir": "www/shared/strftime/"},
            script={"src": "strftime-min.js"},
        ),
    ]


def datepicker_deps(include_css: bool = True) -> HTMLDependency:
    return HTMLDependency(
        name="bootstrap-datepicker",
        version="1.9.0",
        source={"package": "shiny", "subdir": "www/shared/datepicker/"},
        # This CSS is rendered against default Bootstrap
        stylesheet=(
            {"href": "css/bootstrap-datepicker3.min.css"} if include_css else None
        ),
        script={"src": "js/bootstrap-datepicker.min.js"},
        # Need to enable noConflict mode. See #1346.
        head=HTML(
            "<script>(function() { var datepicker = $.fn.datepicker.noConflict(); $.fn.bsDatepicker = datepicker; })();</script>"
        ),
    )


def selectize_deps(include_css: bool = True) -> HTMLDependency:
    return HTMLDependency(
        name="selectize",
        version="0.12.6",
        source={"package": "shiny", "subdir": "www/shared/selectize/"},
        script=[
            {"src": "js/selectize.min.js"},
            {"src": "accessibility/js/selectize-plugin-a11y.min.js"},
        ],
        # This CSS is rendered against default Bootstrap
        stylesheet={"href": "css/selectize.min.css"} if include_css else None,
    )


def jqui_deps() -> HTMLDependency:
    return HTMLDependency(
        name="jquery-ui",
        version="1.12.1",
        source={"package": "shiny", "subdir": "www/shared/jqueryui/"},
        script={"src": "jquery-ui.min.js"},
        stylesheet={"href": "jquery-ui.min.css"},
    )



from __future__ import annotations

__all__ = ("notification_show", "notification_remove")

from typing import TYPE_CHECKING, Any, Literal, Optional

from htmltools import TagChild

from .._docstring import add_example, no_example
from .._utils import rand_hex
from ..session import require_active_session

if TYPE_CHECKING:
    from ..session import Session


@add_example()
def notification_show(
    ui: TagChild,
    *,
    action: Optional[TagChild] = None,
    duration: Optional[int | float] = 5,
    close_button: bool = True,
    id: Optional[str] = None,
    type: Literal["default", "message", "warning", "error"] = "default",
    session: Optional[Session] = None,
) -> str:
    """
    Show a notification to the user.

    A notification is a message that appears near the bottom corner of the app.
    Notifications normally disappear after a short period of time, and should multiple
    notifications appear together, they will stack on top of one another.

    Parameters
    ----------
    ui
        Contents of the notification message.
    action
        Message content that represents an action. For example, this could be a link
        that the user can click on. This is separate from ui so customized layouts can
        handle the main notification content separately from the action content.
    duration
        Number of seconds to display the message before it disappears. Use ``None`` to
        prevent the message from disappearing automatically. The user will need to click
        the corner of the notification to close it.
    close_button
        If ``True``, display a button which will make the notification disappear when
        clicked. If ``False`` do not display.
    id
        An optional unique identifier for the notification. If supplied, any existing
        notification with the same ``id`` will be replaced with this one (otherwise, a
        new notification is created).
    type
        A string which controls the color of the notification. This should be one of
        "default" (gray), "message" (blue), "warning" (yellow), or "error" (red).
    session
        The :class:`~shiny.Session` in which the notification should appear.  If not
        provided, the session is inferred via :func:`~shiny.session.get_current_session`.

    Returns
    -------
    :
        The notification's ``id``.

    See Also
    --------
    * :func:`~shiny.ui.notification_remove`
    * :func:`~shiny.ui.modal`
    """

    session = require_active_session(session)

    ui_ = session._process_ui(ui)
    action_ = session._process_ui(action)

    id = id if id else rand_hex(8)

    payload: dict[str, Any] = {
        "html": ui_["html"],
        "action": action_["html"],
        "deps": ui_["deps"] + action_["deps"],
        "closeButton": close_button,
        "id": id,
        "type": type,
    }

    if duration is None:
        payload.update({"duration": None})
    elif duration:
        payload.update({"duration": duration * 1000})

    session._send_message_sync({"notification": {"type": "show", "message": payload}})

    return id


@no_example()
def notification_remove(id: str, *, session: Optional[Session] = None) -> str:
    """
    Remove a notification.

    :func:`~shiny.ui.notification_remove` provides a way to remove a notification programmatically.
    Notifications can also be removed manually by the user, or automatically after a
    specififed amont of time passes.

    Parameters
    ----------
    id
        The ``id`` of the notification to remove.
    session
        The :class:`~shiny.Session` in which the notification appears. If not provided, the session is inferred via
        :func:`~shiny.session.get_current_session`.

    Returns
    -------
    :
        The notification's ``id``.

    See Also
    --------
    * :func:`~shiny.ui.notification_show`
    * :func:`~shiny.ui.modal`

    Example
    -------
    See :func:`shiny.ui.notification_show`.
    """
    session = require_active_session(session)
    session._send_message_sync({"notification": {"type": "remove", "message": id}})
    return id



import asyncio
import io
import os
from datetime import date

import matplotlib.pyplot as plt
import numpy as np

from shiny.express import render, ui

ui.page_opts(title="Various download examples")

with ui.accordion(open=True):
    with ui.accordion_panel("Simple case"):
        ui.markdown("Downloads a pre-existing file, using its existing name on disk.")

        @render.download(label="Download CSV")
        def download1():
            """
            This is the simplest case. The implementation simply returns the name of a file.
            Note that the function name (`download1`) determines which download_button()
            corresponds to this function.
            """

            path = os.path.join(os.path.dirname(__file__), "mtcars.csv")
            return path

    with ui.accordion_panel("Dynamic data generation"):
        ui.markdown("Downloads a PNG that's generated on the fly.")

        ui.input_text("title", "Plot title", "Random scatter plot")
        ui.input_slider("num_points", "Number of data points", min=1, max=100, value=50)

        @render.download(label="Download plot", filename="image.png")
        def download2():
            """
            Another way to implement a file download is by yielding bytes; either all at
            once, like in this case, or by yielding multiple times. When using this
            approach, you should pass a filename argument to @render.download, which
            determines what the browser will name the downloaded file.
            """

            print(input.num_points())
            x = np.random.uniform(size=input.num_points())
            y = np.random.uniform(size=input.num_points())
            plt.figure()
            plt.scatter(x, y)
            plt.title(input.title())
            with io.BytesIO() as buf:
                plt.savefig(buf, format="png")
                yield buf.getvalue()

    with ui.accordion_panel("Dynamic filename"):
        ui.markdown(
            "Demonstrates that filenames can be generated on the fly (and use Unicode characters!)."
        )

        @render.download(
            label="Download filename",
            filename=lambda: f"新型-{date.today().isoformat()}-{np.random.randint(100, 999)}.csv",
        )
        async def download3():
            await asyncio.sleep(0.25)
            yield "one,two,three\n"
            yield "新,1,2\n"
            yield "型,4,5\n"

    with ui.accordion_panel("Failed downloads"):
        ui.markdown(
            "Throws an error in the download handler, download should not succeed."
        )

        @render.download(label="Download", filename="failuretest.txt")
        async def download4():
            yield "hello"
            raise Exception("This error was caused intentionally")



from htmltools import HTML

from shiny.ui import markdown


def test_markdown():
    assert markdown("# a top level") == HTML("<h1>a top level</h1>\n")
    assert markdown("## a subheading") == HTML("<h2>a subheading</h2>\n")
    assert markdown("[rstudio](https://rstudio.com)") == HTML(
        '<p><a href="https://rstudio.com">rstudio</a></p>\n'
    )

    assert markdown("a ~~paragraph~~ with a link: https://example.com") == HTML(
        '<p>a <s>paragraph</s> with a link: <a href="https://example.com">https://example.com</a></p>\n'
    )

    assert (
        markdown(
            """
        # Hello World

        This is **markdown** and here is some `code`:

        ```python
        print('Hello world!')
        ```
        """
        )
        == HTML(
            "<h1>Hello World</h1>\n<p>This is <strong>markdown</strong> and here is some <code>code</code>:</p>\n<pre><code class=\"language-python\">print('Hello world!')\n</code></pre>\n"
        )
    )

    assert (
        markdown(
            """
        # Hello World

        This is **markdown** and here is some `code`:

            print('Hello world!')
        """
        )
        == HTML(
            "<h1>Hello World</h1>\n<p>This is <strong>markdown</strong> and here is some <code>code</code>:</p>\n<pre><code>print('Hello world!')\n</code></pre>\n"
        )
    )



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.tags.p("The checkbox group controls the select input"),
    ui.input_checkbox_group(
        "inCheckboxGroup", "Input checkbox", ["Item A", "Item B", "Item C"]
    ),
    ui.input_select("inSelect", "Select input", ["Item A", "Item B", "Item C"]),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    def _():
        x = input.inCheckboxGroup()

        # Can use [] to remove all choices
        if x is None:
            x = []
        elif isinstance(x, str):
            x = [x]

        ui.update_select(
            "inSelect",
            label="Select input label " + str(len(x)),
            choices=x,
            selected=x[len(x) - 1] if len(x) > 0 else None,
        )


app = App(app_ui, server)



from __future__ import annotations

__all__ = ("include_js", "include_css")

import glob
import hashlib
import os
import shutil
import tempfile
from pathlib import Path
from typing import Literal

# TODO: maybe these include_*() functions should actually live in htmltools?
from htmltools import HTMLDependency, Tag, TagAttrValue, tags

from .._docstring import add_example

# TODO: it's bummer that, when method="link_files" and path is in the same directory
# as the app, the app's source will be included. Should we just not copy .py/.r files?


@add_example()
def include_js(
    path: Path | str,
    *,
    method: Literal["link", "link_files", "inline"] = "link",
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Include a JavaScript file.

    Parameters
    ----------
    path
        A path to a JS file.
    method
        One of the following:

        * ``"link"`` is the link to the CSS file via a :func:`~shiny.ui.tags.link` tag. This
          method is generally preferable to ``"inline"`` since it allows the browser to
          cache the file.
        * ``"link_files"`` is the same as ``"link"``, but also allow for the CSS file to
          request other files within ``path``'s immediate parent directory (e.g.,
          ``@import()`` another file). Note that this isn't the default behavior because
          you should **be careful not to include files in the same directory as ``path``
          that contain sensitive information**. A good general rule of thumb to follow
          is to have ``path`` be located in a subdirectory of the app directory. For
          example, if the app's source is located at ``/app/app.py``, then ``path``
          should be somewhere like ``/app/css/custom.css`` (and all the other relevant
          accompanying 'safe' files should be located under ``/app/css/``).
        * ``"inline"`` is the inline the CSS file contents within a
          :func:`~shiny.ui.tags.style` tag.
    **kwargs
        Attributes which are passed on to `~shiny.ui.tags.script`.


    Returns
    -------
    :
        A :func:`~shiny.ui.tags.script` tag.

    Note
    ----
    This places a :func:`~shiny.ui.tags.script` tag in the :func:`~shiny.ui.tags.body` of the
    document. If you want to place the tag in the :func:`~shiny.ui.tags.head` of the
    document instead, you can wrap it in ``head_content`` (in this case, just
    make sure you're aware that the DOM probably won't be ready when the script
    is executed).

    ```{python}
    #| eval: false
    ui.page_fluid(
        ui.head_content(ui.include_js("custom.js")),
    )

    # Alternately you can inline Javscript by changing the method.
    ui.page_fluid(
        ui.head_content(ui.include_js("custom.js", method = "inline")),
    )
    ```

    See Also
    --------
    * :func:`~shiny.ui.tags.script`
    * :func:`~shiny.ui.include_css`
    """
    file_path = check_path(path)

    if method == "inline":
        return tags.script(read_utf8(file_path), **kwargs)

    include_files = method == "link_files"
    path_dest, hash = maybe_copy_files(file_path, include_files)

    dep, src = create_include_dependency("include-js-" + hash, path_dest, include_files)

    return tags.script(dep, src=src, **kwargs)


@add_example()
def include_css(
    path: Path | str, *, method: Literal["link", "link_files", "inline"] = "link"
) -> Tag:
    """
    Include a CSS file.

    Parameters
    ----------
    path
        A path to a CSS file.
    method
        One of the following:

        * ``"link"`` is the link to the CSS file via a :func:`~shiny.ui.tags.link` tag. This
          method is generally preferable to ``"inline"`` since it allows the browser to
          cache the file.
        * ``"link_files"`` is the same as ``"link"``, but also allow for the CSS file to
          request other files within ``path``'s immediate parent directory (e.g.,
          ``@import()`` another file). Note that this isn't the default behavior because
          you should **be careful not to include files in the same directory as ``path``
          that contain sensitive information**. A good general rule of thumb to follow
          is to have ``path`` be located in a subdirectory of the app directory. For
          example, if the app's source is located at ``/app/app.py``, then ``path``
          should be somewhere like ``/app/css/custom.css`` (and all the other relevant
          accompanying 'safe' files should be located under ``/app/css/``).
        * ``"inline"`` is the inline the CSS file contents within a
          :func:`~shiny.ui.tags.style` tag.


    Returns
    -------
    :

        If ``method="inline"``, returns a :func:`~shiny.ui.tags.style` tag; otherwise, returns a
        :func:`~shiny.ui.tags.link` tag.

    Note
    ----
    By default this places a :func:`~shiny.ui.tags.link` (or :func:`~shiny.ui.tags.style`) tag in
    the :func:`~shiny.ui.tags.body` of the document, which isn't optimal for performance, and
    may result in a Flash of Unstyled Content (FOUC). To instead place the CSS in the
    :func:`~shiny.ui.tags.head` of the document, you can wrap it in ``head_content``:

    ```{python}
    #| eval: false
    from htmltools import head_content
    from shiny import ui

    ui.page_fluid(
        ui.head_content(ui.include_css("custom.css")),

        # You can also inline css by passing a dictionary with a `style` element.
        ui.div(
            {"style": "font-weight: bold;"},
            ui.p("Some text!"),
        )
    )
    ```

    See Also
    --------
    * :func:`~shiny.ui.tags.style`
    * :func:`~shiny.ui.tags.link`
    * :func:`~shiny.ui.include_js`
    """

    file_path = check_path(path)
    if method == "inline":
        return tags.style(read_utf8(file_path), type="text/css")

    include_files = method == "link_files"
    path_dest, hash = maybe_copy_files(file_path, include_files)

    dep, src = create_include_dependency(
        "include-css-" + hash, path_dest, include_files
    )

    return tags.link(dep, href=src, rel="stylesheet")


# ---------------------------------------------------------------------------
# Include helpers
# ---------------------------------------------------------------------------


def check_path(path: Path | str) -> Path:
    path = Path(path)
    if not path.exists():
        err = f"""
        {path.absolute()} does not exist.
        Files are typically placed in the app directory and refered to with 'Path(__file__) / {path.name}'
        """
        raise RuntimeError(err)
    return path


def create_include_dependency(
    name: str, path: str, include_files: bool
) -> tuple[HTMLDependency, str]:
    dep = HTMLDependency(
        name,
        DEFAULT_VERSION,
        source={"subdir": os.path.dirname(path)},
        all_files=include_files,
    )

    # source_path_map() tells us where the source subdir is mapped to on the client
    # (i.e., session._register_web_dependency() uses the same thing to determine where
    # to mount the subdir, but we can't assume an active session at this point).
    src = os.path.join(dep.source_path_map()["href"], os.path.basename(path))

    return dep, src


def maybe_copy_files(path: Path | str, include_files: bool) -> tuple[str, str]:
    hash = get_hash(path, include_files)

    # To avoid unnecessary work when the same file is included multiple times,
    # use a directory scoped by a hash of the file.
    tmpdir = os.path.join(tempfile.gettempdir(), "shiny_include_files", hash)
    path_dest = os.path.join(tmpdir, os.path.basename(path))

    # Since the hash/tmpdir should represent all the files in the path's directory,
    # we can simply return here
    if os.path.exists(path_dest):
        return path_dest, hash

    # Otherwise, make sure we have a clean slate
    if os.path.exists(tmpdir):
        shutil.rmtree(tmpdir)

    if include_files:
        shutil.copytree(os.path.dirname(path), tmpdir)
    else:
        os.makedirs(tmpdir, exist_ok=True)
        shutil.copy(path, path_dest)

    return path_dest, hash


def get_hash(path: Path | str, include_files: bool) -> str:
    if include_files:
        key = get_file_key(path)
    else:
        dir = os.path.dirname(path)
        files = glob.iglob(os.path.join(dir, "**"), recursive=True)
        key = "\n".join([get_file_key(x) for x in files])
    return hash_deterministic(key)


def get_file_key(path: Path | str) -> str:
    path = Path(path)
    return str(path) + "-" + str(path.stat().st_mtime)


def hash_deterministic(s: str) -> str:
    """
    Returns a deterministic hash of the given string.
    """
    return hashlib.sha1(s.encode("utf-8")).hexdigest()


def read_utf8(path: Path | str) -> str:
    with open(path, "r", encoding="utf-8") as f:
        return f.read()


DEFAULT_VERSION = "0.0"



from __future__ import annotations

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_row_selection(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    # The purpose of this test is to make sure that the data grid can work on Pandas
    # data frames that use an index that is not simply 0-based integers.

    grid = controller.OutputDataFrame(page, "grid")
    detail = controller.OutputDataFrame(page, "detail")
    selected_rows = controller.OutputCode(page, "selected_rows")

    grid.expect_cell("three", row=2, col=0)
    detail.expect_nrow(0)
    detail.expect_ncol(3)
    selected_rows.expect_value("()")

    grid.select_rows([2])
    detail.expect_nrow(1)
    detail.expect_cell("three", row=0, col=0)
    selected_rows.expect_value("(2,)")

    # Ensure that keys are in sorted order, not the order in which they were selected
    # row1.click(modifiers=["Shift"])
    grid.loc_body.locator("td").nth(0).click(modifiers=["Shift"])
    detail.expect_nrow(3)
    selected_rows.expect_value("(0, 1, 2)")



from __future__ import annotations

__all__ = ("output_data_frame",)

from htmltools import Tag

from ..._docstring import add_example
from ..._namespaces import resolve_id
from .._html_deps_py_shiny import data_frame_deps
from ..fill import as_fill_item, as_fillable_container


@add_example(ex_dir="../../api-examples/data_frame")
def output_data_frame(id: str) -> Tag:
    """
    Create an output container for an interactive table or grid. Features fast
    virtualized scrolling, sorting, filtering, and row selection (single or multiple).

    Parameters
    ----------
    id
        An output id.

    Returns
    -------
    :
        A UI element.

    See Also
    --------
    * :class:`shiny.render.data_frame`
    """
    return as_fillable_container(
        as_fill_item(
            Tag(
                "shiny-data-frame",
                data_frame_deps(),
                id=resolve_id(id),
            ),
        )
    )



from shiny import reactive
from shiny.express import input, ui

ui.input_action_button("btn_show", "Show tooltip", class_="mt-3 me-3")
ui.input_action_button("btn_close", "Close tooltip", class_="mt-3 me-3")
ui.input_action_button(
    "btn_update", "Update tooltip phrase (and show tooltip)", class_="mt-3 me-3"
)

with ui.tooltip(id="tooltip_id"):
    ui.input_action_button(
        "btn_w_tooltip",
        "A button w/ a tooltip",
        class_="btn-primary mt-5",
    )
    "A message"


@reactive.effect
@reactive.event(input.btn_show)
def _():
    ui.update_tooltip("tooltip_id", show=True)


@reactive.effect
@reactive.event(input.btn_close)
def _():
    ui.update_tooltip("tooltip_id", show=False)


@reactive.effect
@reactive.event(input.btn_update)
def _():
    content = "A " + " ".join(["NEW" for _ in range(input.btn_update())]) + " message"

    ui.update_tooltip("tooltip_id", content, show=True)


@reactive.effect
@reactive.event(input.btn_w_tooltip)
def _():
    ui.notification_show("Button clicked!", duration=3, type="message")



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("input_switch")


def test_input_switch_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    somevalue = controller.InputSwitch(page, "somevalue")

    expect(somevalue.loc_label).to_have_text("Some value")
    somevalue.expect_label("Some value")

    somevalue.expect_checked(False)
    somevalue.expect_width(None)

    expect(controller.OutputUi(page, "value").loc).to_have_text("False")

    somevalue.set(True)

    somevalue.expect_checked(True)

    somevalue.set(False)
    somevalue.expect_checked(False)

    somevalue.set(True)
    somevalue.expect_checked(True)

    expect(controller.OutputUi(page, "value").loc).to_have_text("True")



import pandas as pd

from shiny import App, Inputs, reactive, render, ui

app_ui = ui.page_fluid(
    ui.markdown(
        """
    ## Description

    When you add a row, click on it and click the clear button you get:


    """
    ),
    ui.input_action_button("add_row", "Add row"),
    ui.input_action_button("clear_table", "Clear table"),
    ui.output_text_verbatim("number_of_selected_rows"),
    ui.output_data_frame("df1"),
)


def server(input: Inputs):
    df = reactive.Value(pd.DataFrame(columns=["A", "B"]))

    @render.data_frame
    def df1():
        return render.DataGrid(df(), selection_mode="rows")

    @reactive.effect
    @reactive.event(input.add_row)
    def _():
        old_df = df()
        new_df = pd.concat(  # pyright: ignore[reportUnknownMemberType]
            [old_df, pd.DataFrame([[1, 2]], columns=["A", "B"])]
        )
        df.set(new_df)

    @render.text
    def number_of_selected_rows():
        df_selected = df1.data_view(selected=True)
        return f"Selected rows: {len(df_selected)}"

    @reactive.effect
    @reactive.event(input.clear_table)
    def _():
        df.set(pd.DataFrame(columns=["A", "B"]))


app = App(app_ui, server)



from shiny import App, Inputs, Outputs, Session, render, ui

states = {
    "East Coast": {"NY": "New York", "NJ": "New Jersey", "CT": "Connecticut"},
    "West Coast": {"WA": "Washington", "OR": "Oregon", "CA": "California"},
    "Midwest": {"MN": "Minnesota", "WI": "Wisconsin", "IA": "Iowa"},
}

state_without_groups = {"NY": "New York", "NJ": "New Jersey", "CT": "Connecticut"}
state_without_keys = ["New York", "New Jersey", "Connecticut"]

app_ui = ui.page_fluid(
    ui.input_selectize(
        "state1",
        "Choose a state:",
        states,
        multiple=True,
    ),
    ui.input_selectize(
        "state2",
        "Selectize Options",
        states,
        multiple=True,
        options=(
            {
                "placeholder": "Enter text",
                "render": ui.js_eval(
                    '{option: function(item, escape) {return "<div><strong>Select " + escape(item.label) + "</strong></div>";}}'
                ),
                "create": True,
            }
        ),
    ),
    ui.input_selectize(
        "state3",
        "Single Selectize",
        state_without_groups,
        multiple=False,
        options={"plugins": ["clear_button"]},
    ),
    ui.input_selectize(
        "state4",
        "Simple Selectize",
        state_without_keys,
        multiple=False,
    ),
    ui.hr(),
    ui.output_code("value1"),
    ui.output_code("value2"),
    ui.output_code("value3"),
    ui.output_code("value4"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.code
    def value1():
        return str(input.state1())

    @render.code
    def value2():
        return str(input.state2())

    @render.code
    def value3():
        return str(input.state3())

    @render.code
    def value4():
        return str(input.state4())


app = App(app_ui, server)



from __future__ import annotations

import ast
import functools
import inspect
import linecache
import sys
import types
from typing import (
    Any,
    Callable,
    Dict,
    Protocol,
    TypeVar,
    cast,
    overload,
    runtime_checkable,
)

from ..._docstring import no_example
from ..._shinyenv import is_pyodide
from ._func_displayhook import _expressify_decorator_function_def
from ._helpers import find_code_for_func
from ._node_transformers import (
    DisplayFuncsTransformer,
    FuncBodyDisplayHookTransformer,
    TargetFunctionTransformer,
    expressify_decorator_func_name,
    sys_alias,
)

# It's quite expensive to decorate with expressify, and it could be done to inner
# functions where the outer function is called a lot. Use a cache to save us from having
# to do the expensive stuff (parsing, transforming, compiling) more than once.
code_cache: Dict[types.CodeType, types.CodeType] = {}

T = TypeVar("T")
TFunc = TypeVar("TFunc", bound=Callable[..., Any])


def auto_displayhook(x: T) -> T:
    if x is not None:
        sys.displayhook(x)
    return x


@runtime_checkable
class WrappedFunction(Protocol):
    __wrapped__: types.FunctionType


def unwrap(fn: TFunc) -> TFunc:
    while isinstance(fn, WrappedFunction):
        fn = fn.__wrapped__
    return fn


expressify_attr = "__expressify__"


def expressify_unwrap_inplace() -> Callable[[TFunc], TFunc]:
    """
    Like `expressify`, but far more violent. This will attempt to traverse any
    decorators between this one and the function, and then modify the function _in
    place_. It will then return the function that was passed in.
    """

    def decorator(fn: TFunc) -> TFunc:
        unwrapped_fn = unwrap(fn)

        # Check if we've already done this
        if hasattr(unwrapped_fn, expressify_attr):
            return fn

        if is_pyodide:
            # Disable code caching on Pyodide due to bug in hashing bytecode in 0.22.1.
            # When Pyodide is updated to a newer version, this will be not be needed.
            # https://github.com/posit-dev/py-shiny/issues/1042#issuecomment-1901945787
            fcode = _transform_body(cast(types.FunctionType, unwrapped_fn))
        else:
            if unwrapped_fn.__code__ in code_cache:
                fcode = code_cache[unwrapped_fn.__code__]
            else:
                # Save for next time
                fcode = _transform_body(cast(types.FunctionType, unwrapped_fn))
                code_cache[unwrapped_fn.__code__] = fcode

        unwrapped_fn.__code__ = fcode
        setattr(unwrapped_fn, expressify_attr, True)
        return fn

    return decorator


@overload
def expressify(fn: TFunc) -> TFunc: ...


@overload
def expressify(*, has_docstring: bool = False) -> Callable[[TFunc], TFunc]: ...


@no_example()
def expressify(
    fn: TFunc | None = None,
    *,
    has_docstring: bool = False,
) -> TFunc | Callable[[TFunc], TFunc]:
    """
    Decorate a function so that output is captured as in Shiny Express

    In a Shiny Express app, the output of each line of the app file is captured and
    displayed in the UI. However, if the app calls a function, only the return value of
    the function is displayed. This decorator changes the behavior of the function so
    that when it is executed, the result of each line is captured and displayed, just
    like code at the top level of a Shiny Express app.

    Parameters
    ----------
    fn :
        The function to decorate. If not provided, this is a decorator factory.
    has_docstring :
        Whether the function has a docstring. Set this to `True` if the function to
        decorate has a docstring. This tells `expressify()` to *not* capture the
        docstring and display it in the UI.

    Returns
    -------
    :
        A function that returns `None`, or a decorator for a function that returns
        `None`.

    See Also
    --------
    * ~shiny.render.express
    * ~shiny.express.ui.hold
    """

    def decorator(fn: TFunc) -> TFunc:
        if is_pyodide:
            # Disable code caching on Pyodide due to bug in hashing bytecode in 0.22.1.
            # When Pyodide is updated to a newer version, this will be not be needed.
            # https://github.com/posit-dev/py-shiny/issues/1042#issuecomment-1901945787
            fcode = _transform_body(cast(types.FunctionType, fn), has_docstring)
        else:
            if fn.__code__ in code_cache:
                fcode = code_cache[fn.__code__]
            else:
                # Save for next time
                fcode = _transform_body(cast(types.FunctionType, fn), has_docstring)
                code_cache[fn.__code__] = fcode

        # Create a new function from the code object
        new_func = types.FunctionType(
            code=fcode,
            # We add calls to sys.displayhook, but we use our own alias for
            # `sys` so we don't have to worry about whether the user happens to
            # have a different `sys` alias in scope.
            globals={
                sys_alias: auto_displayhook,
                expressify_decorator_func_name: _expressify_decorator_function_def,
                **fn.__globals__,
            },
            name=fn.__name__,
            argdefs=fn.__defaults__,
            closure=fn.__closure__,
        )
        # Need to copy over some attributes, for some reason FunctionType()
        # doesn't include all of these
        new_func.__kwdefaults__ = fn.__kwdefaults__
        new_func.__dict__.update(fn.__dict__)
        return cast(TFunc, functools.wraps(fn)(new_func))

    if fn is not None:
        return decorator(fn)

    return decorator


def _transform_body(
    fn: types.FunctionType,
    has_docstring: bool = False,
) -> types.CodeType:
    # The approach we take here is much more complicated than what you'd expect.
    #
    # The simple approach is to use ast.parse() to get an AST for the function,
    # use an ast.NodeTransformer to modify the AST, and then compile() and
    # exec() each top-level node. However, this approach does not work correctly
    # with closures, because compile/exec do not support scopes besides global
    # and local.
    #
    # Instead, we need to ast.parse(), transform just the part of the AST that
    # is the function we care about, and compile the entire module, and then
    # find the code object for the function we want. We then use that code
    # object to programmatically create a new function.

    filename = inspect.getsourcefile(fn)
    if filename is None:
        raise RuntimeError(
            f"Failed to find source code for function '{fn.__name__}'."
            " This should never happen, please file an issue!"
        )

    parsed_ast = read_ast(filename)
    if parsed_ast is None:
        raise RuntimeError(
            f"Failed to read source code for function '{fn.__name__}'."
            " This should never happen, please file an issue!"
        )

    # A wrapper for _transform_function_ast that conveys the value of has_docstring.
    def transform_function_ast_local(node: ast.AST) -> ast.AST:
        return _transform_function_ast(node, has_docstring)

    tft = TargetFunctionTransformer(
        fn,
        # If we find `fn` in the AST, use transform its body to use displayhook
        transform_function_ast_local,
    )

    new_ast = tft.visit(parsed_ast)
    if not tft.found:
        raise RuntimeError(
            f"Failed to find function '{fn.__name__}' in AST."
            " This should never happen, please file an issue!"
        )

    # The new AST contains new nodes; give them locations so the compiler will
    # accept them
    new_ast = ast.fix_missing_locations(new_ast)

    # Compile the new AST into a code object
    compiled_code = compile(
        new_ast,
        filename=filename,
        mode="exec",
    )

    fcode = find_code_for_func(
        compiled_code, fn, compare_decorated_code_objects(tft.found)
    )
    if fcode is None:
        raise RuntimeError(
            f"Failed to find code object for function '{fn.__name__}'."
            " This should never happen, please file an issue!"
        )

    return fcode


def read_ast(filename: str) -> ast.Module | None:
    # This is the logic we originally used to read the AST, but it doesn't work when the
    # function being decorated is defined in a Jupyter notebook cell. However, the code
    # is available in linecache (without which, tracebacks wouldn't work right).

    # with open(filename) as fd:
    #     return ast.parse(fd.read(), filename=filename)

    linecache.checkcache(filename)
    lines = linecache.getlines(filename)
    if len(lines) == 0:
        return None
    return ast.parse("".join(lines), filename=filename)


def _transform_function_ast(node: ast.AST, has_docstring: bool = False) -> ast.AST:
    if not isinstance(node, ast.FunctionDef):
        return node
    func_node = cast(
        ast.FunctionDef,
        FuncBodyDisplayHookTransformer(has_docstring).visit(node),
    )
    func_node.body = [
        DisplayFuncsTransformer().visit(child) for child in func_node.body
    ]
    return func_node


def compare_decorated_code_objects(
    func_ast: ast.FunctionDef,
) -> Callable[[types.CodeType, types.CodeType], bool]:
    linenos = [*[x.lineno for x in func_ast.decorator_list], func_ast.lineno]

    def comparator(candidate: types.CodeType, target: types.CodeType) -> bool:
        if candidate.co_name == target.co_name:
            if candidate.co_firstlineno == target.co_firstlineno:
                return True
            elif candidate.co_firstlineno in linenos:
                # Ordinarily we'd just compare co_name (or co_qualname if possible) and
                # co_firstlineno for equality, but sometimes the co_firstlineno points
                # to the first decorator (most cases) and other times it points to the
                # def line (when compiling node-by-node, as in express mode).
                return True

        return False

    return comparator


__builtins__[sys_alias] = auto_displayhook
__builtins__[expressify_decorator_func_name] = _expressify_decorator_function_def



from shiny.express import input, render, ui

with ui.navset_underline(id="selected_navset_underline"):
    with ui.nav_panel("Tab 1"):
        "Tab 1 content"
    ui.nav_spacer()
    with ui.nav_panel("Tab 2"):
        "Tab 2 content"
    with ui.nav_panel("Tab 3"):
        "Tab 3 content"
ui.h5("Selected:")


@render.code
def _():
    return input.selected_navset_underline()



from shiny.express import ui


def navset_sidebar():
    from shiny import ui as core_ui

    return core_ui.sidebar(core_ui.markdown("Sidebar content"))


ui.page_opts(id="page_navbar_sidebar", sidebar=navset_sidebar())

with ui.nav_panel("Data"):
    "This page could be used to pick a dataset."

with ui.nav_panel("View"):
    "This page could be used to view the dataset."



import seaborn as sns

# Import data from shared.py
from shared import df

from shiny import App, render, ui

# The contents of the first 'page' is a navset with two 'panels'.
page1 = ui.navset_card_underline(
    ui.nav_panel("Plot", ui.output_plot("hist")),
    ui.nav_panel("Table", ui.output_data_frame("data")),
    footer=ui.input_select(
        "var", "Select variable", choices=["bill_length_mm", "body_mass_g"]
    ),
    title="Penguins data",
)

app_ui = ui.page_navbar(
    ui.nav_spacer(),  # Push the navbar items to the right
    ui.nav_panel("Page 1", page1),
    ui.nav_panel("Page 2", "This is the second 'page'."),
    title="Shiny navigation components",
)


def server(input, output, session):
    @render.plot
    def hist():
        p = sns.histplot(df, x=input.var(), facecolor="#007bc2", edgecolor="white")
        return p.set(xlabel=None)

    @render.data_frame
    def data():
        return df[["species", "island", input.var()]]


app = App(app_ui, server)



from shiny import render
from shiny.express import input, ui

with ui.accordion(id="express_accordion", open=["Panel 1", "Panel 2"]):
    with ui.accordion_panel("Panel 1"):
        ui.input_slider("n", "N", 1, 100, 50)

    with ui.accordion_panel("Panel 2"):

        @render.text
        def txt():
            return f"n = {input.n()}"



# pyright: reportUnknownMemberType=false

from conftest import create_example_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc
from shiny.types import MISSING

inputs_update_app = create_example_fixture("inputs-update")


def check_vals(page: Page, label: str, num: int):
    opt_labels = [f"option label {num} {type}" for type in ["A", "B"]]
    opt_vals = [f"option-{num}-{type}" for type in ["A", "B"]]

    text = controller.InputText(page, "inText")
    text.expect_label(f"New {label}")
    text.expect_value(f"New text {num}")

    innum = controller.InputNumeric(page, "inNumber")
    innum.expect_label("Number input:")
    innum.expect_value(str(num))

    innum2 = controller.InputNumeric(page, "inNumber2")
    innum2.expect_label(f"Number {label}")
    innum2.expect_value(str(num))

    slider = controller.InputSlider(page, "inSlider")
    slider.expect_label(f"Slider {label}")
    slider.expect_value(str(num))

    slider2 = controller.InputSliderRange(page, "inSlider2")
    slider2.expect_label("Slider input 2:")
    slider2.expect_value((str(num - 1), str(num + 1)))

    slider3 = controller.InputSliderRange(page, "inSlider3")
    slider3.expect_value((MISSING, str(num + 2)))
    slider3.expect_label("Slider input 3:")

    date = controller.InputDate(page, "inDate")
    date.expect_label(f"Date {label}")
    date.expect_value(f"2013-04-{num:02}")

    date_range_input = controller.InputDateRange(page, "inDateRange")
    date_range_input.expect_label(f"Date range {label}")
    date_range_input.date_start.expect_value(f"2013-01-{num:02}")
    date_range_input.date_end.expect_value(f"2013-12-{num:02}")
    # TODO-future; Check min/max date
    # min=date(2001, 1, num),
    # max=date(2030, 1, num),

    cb = controller.InputCheckbox(page, "inCheckbox")
    cb.expect_label("Checkbox input")
    cb.expect_checked((num % 2) == 1)

    cb_grp_input = controller.InputCheckboxGroup(page, "inCheckboxGroup")
    cb_grp_input.expect_label(f"Checkbox group {label}")
    cb_grp_input.expect_choices(opt_vals)
    cb_grp_input.expect_choice_labels(opt_labels)
    cb_grp_input.expect_selected([opt_vals[0]])

    radio_btns_input = controller.InputRadioButtons(page, "inRadio")
    radio_btns_input.expect_label(f"Radio {label}")
    radio_btns_input.expect_choice_labels(opt_labels)
    radio_btns_input.expect_choices(opt_vals)
    radio_btns_input.expect_selected(opt_vals[0])

    select = controller.InputSelect(page, "inSelect")
    select.expect_label(f"Select {label}")
    select.expect_choice_labels(opt_labels)
    select.expect_choices(opt_vals)
    select.expect_selected(opt_vals[0])

    selectize = controller.InputSelect(page, "inSelect2")
    selectize.expect_label(f"Select label {label}")
    selectize.expect_choices(opt_vals)
    selectize.expect_selected(opt_vals[1])

    # TODO-future: proper tab output classes
    panel1_is_visible = (num % 2) == 0
    panel1_expect = expect(page.locator('div[data-value="panel1"]'))
    panel2_expect = expect(page.locator('div[data-value="panel2"]'))
    if panel1_is_visible:
        panel1_expect.to_be_visible()
        panel2_expect.not_to_be_visible()
        panel1_expect.to_have_text("This is the first panel.")
    else:
        panel1_expect.not_to_be_visible()
        panel2_expect.to_be_visible()
        panel2_expect.to_have_text("This is the second panel.")


def test_updates(page: Page, inputs_update_app: ShinyAppProc):
    page.goto(inputs_update_app.url)

    for n in [9, 10]:
        label = f"LABEL {n} TEXT"
        controller.InputText(page, "control_label").set(label)
        controller.InputSlider(page, "control_num").set(str(n))
        check_vals(page, label, n)



from shiny import render, ui
from shiny.express import output_args
from shiny.types import MISSING


def test_decorator_plot_sizing():
    """render.plot width/height are passed through to implicit output"""

    @render.plot(width=1280, height=960)
    def foo(): ...

    rendered = str(foo.tagify())
    assert "1280px" in rendered
    assert "960px" in rendered
    assert rendered == str(ui.output_plot("foo", width=1280, height=960))


def test_decorator_plot_default():
    """render.plot default is the same as ui.output_plot default"""

    @render.plot()
    def foo(): ...

    rendered = str(foo.tagify())
    assert rendered == str(ui.output_plot("foo"))


def test_decorator_output_args():
    """@output_args is respected"""

    @output_args(width="640px", height="480px")
    @render.plot()
    def foo(): ...

    rendered = str(foo.tagify())
    assert rendered == str(ui.output_plot("foo", width="640px", height="480px"))


def test_decorator_output_args_priority():
    """@output_args should override render.plot width/height"""

    @output_args(width="640px", height=480)
    @render.plot(width=1280, height=960)
    def foo(): ...

    rendered = str(foo.tagify())
    # Note "640px" => 640 and 480 => "480px"
    assert rendered == str(ui.output_plot("foo", width=640, height="480px"))


def test_decorator_output_args_MISSING():
    """Not saying we support this, but test how MISSING interacts"""

    @output_args(width=MISSING)
    @render.plot(width=1280, height=MISSING)
    def foo(): ...

    rendered = str(foo.tagify())
    assert rendered == str(ui.output_plot("foo", width="1280px"))



from __future__ import annotations

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild

from ._html_deps_shinyverse import components_dependencies


def web_component(
    tag_name: str,
    *args: TagChild | TagAttrs,
    **kwargs: TagAttrValue,
) -> Tag:
    return Tag(
        tag_name,
        components_dependencies(),
        *args,
        _add_ws=False,
        **kwargs,
    )



from __future__ import annotations

__all__ = ("input_date", "input_date_range")

import json
from datetime import date
from typing import Optional

from htmltools import Tag, TagAttrValue, TagChild, css, div, span, tags

from .._docstring import add_example
from .._namespaces import resolve_id
from ._html_deps_external import datepicker_deps
from ._utils import shiny_input_label


@add_example()
def input_date(
    id: str,
    label: TagChild,
    *,
    value: Optional[date | str] = None,
    min: Optional[date | str] = None,
    max: Optional[date | str] = None,
    format: str = "yyyy-mm-dd",
    startview: str = "month",
    weekstart: int = 0,
    language: str = "en",
    width: Optional[str] = None,
    autoclose: bool = True,
    datesdisabled: Optional[list[str]] = None,
    daysofweekdisabled: Optional[list[int]] = None,
) -> Tag:
    """
    Creates a text input which, when clicked on, brings up a calendar that the user can
    click on to select dates.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        The starting date. Either a :class:`~datetime.date` object, or a string in
        `yyyy-mm-dd` format. If None (the default), will use the current date in the
        client's time zone.
    min
        The minimum allowed date. Either a :class:`~datetime.date` object, or a string in
        yyyy-mm-dd format.
    max
        The maximum allowed date. Either a :class:`~datetime.date` object, or a string in
        yyyy-mm-dd format.
    format
        The format of the date to display in the browser. Defaults to `"yyyy-mm-dd"`.
    startview
        The date range shown when the input object is first clicked. Can be "month" (the
        default), "year", or "decade".
    weekstart
        Which day is the start of the week. Should be an integer from 0 (Sunday) to 6
        (Saturday).
    language
        The language used for month and day names. Default is "en". Other valid values
        include "ar", "az", "bg", "bs", "ca", "cs", "cy", "da", "de", "el", "en-AU",
        "en-GB", "eo", "es", "et", "eu", "fa", "fi", "fo", "fr-CH", "fr", "gl", "he",
        "hr", "hu", "hy", "id", "is", "it-CH", "it", "ja", "ka", "kh", "kk", "ko", "kr",
        "lt", "lv", "me", "mk", "mn", "ms", "nb", "nl-BE", "nl", "no", "pl", "pt-BR",
        "pt", "ro", "rs-latin", "rs", "ru", "sk", "sl", "sq", "sr-latin", "sr", "sv",
        "sw", "th", "tr", "uk", "vi", "zh-CN", and "zh-TW".
    width
        The CSS width, e.g. '400px', or '100%'
    autoclose
        Whether or not to close the datepicker immediately when a date is selected.
    datesdisabled
        Which dates should be disabled (in `yyyy-mm-dd` format).
    daysofweekdisabled
        Days of the week that should be disabled. Should be a integer vector with values
        from 0 (Sunday) to 6 (Saturday).

    Returns
    -------
    :
        A UI element.

    Note
    ----
    The date ``format`` string specifies how the date will be displayed in the browser.
    It allows the following values:

    - ``yy``: Year without century (12)
    - ``yyyy``: Year with century (2012)
    - ``mm``: Month number, with leading zero (01-12)
    - ``m``: Month number, without leading zero (1-12)
    - ``M``: Abbreviated month name
    - ``MM``: Full month name
    - ``dd``: Day of month with leading zero
    - ``d``: Day of month without leading zero
    - ``D``: Abbreviated weekday name
    - ``DD``: Full weekday name

    Notes
    ------
    ::: {.callout-note title="Server value"}
    A :class:`~datetime.date` object.
    :::

    See Also
    --------
    * :func:`~shiny.ui.update_date`
    * :func:`~shiny.ui.input_date_range`
    """

    resolved_id = resolve_id(id)
    return div(
        shiny_input_label(resolved_id, label),
        _date_input_tag(
            id=resolved_id,
            value=value,
            min=min,
            max=max,
            format=format,
            startview=startview,
            weekstart=weekstart,
            language=language,
            autoclose=autoclose,
            data_date_dates_disabled=json.dumps(datesdisabled),
            data_date_days_of_week_disabled=json.dumps(daysofweekdisabled),
        ),
        id=resolved_id,
        class_="shiny-date-input form-group shiny-input-container",
        style=css(width=width),
    )


@add_example()
def input_date_range(
    id: str,
    label: TagChild,
    *,
    start: Optional[date | str] = None,
    end: Optional[date | str] = None,
    min: Optional[date | str] = None,
    max: Optional[date | str] = None,
    format: str = "yyyy-mm-dd",
    startview: str = "month",
    weekstart: int = 0,
    language: str = "en",
    separator: str = " to ",
    width: Optional[str] = None,
    autoclose: bool = True,
) -> Tag:
    """
    Creates a pair of text inputs which, when clicked on, bring up calendars that the
    user can click on to select dates.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    start
        The initial start date. Either a :class:`~datetime.date` object, or a string in
        yyyy-mm-dd format. If ``None`` (the default), will use the current date in the
        client's time zone.
    end
        The initial end date. Either a :class:`~datetime.date` object, or a string in
        yyyy-mm-dd format. If ``None`` (the default), will use the current date in the
        client's time zone.
    min
        The minimum allowed date. Either a :class:`~datetime.date` object, or a string in
        yyyy-mm-dd format.
    max
        The maximum allowed date. Either a :class:`~datetime.date` object, or a string in
        yyyy-mm-dd format.
    format
        The format of the date to display in the browser.
    startview
        The date range shown when the input object is first clicked. Can be "month" (the
        default), "year", or "decade".
    weekstart
        Which day is the start of the week. Should be an integer from 0 (Sunday) to 6
        (Saturday).
    language
        The language used for month and day names. Default is "en". Other valid values
        include "ar", "az", "bg", "bs", "ca", "cs", "cy", "da", "de", "el", "en-AU",
        "en-GB", "eo", "es", "et", "eu", "fa", "fi", "fo", "fr-CH", "fr", "gl", "he",
        "hr", "hu", "hy", "id", "is", "it-CH", "it", "ja", "ka", "kh", "kk", "ko", "kr",
        "lt", "lv", "me", "mk", "mn", "ms", "nb", "nl-BE", "nl", "no", "pl", "pt-BR",
        "pt", "ro", "rs-latin", "rs", "ru", "sk", "sl", "sq", "sr-latin", "sr", "sv",
        "sw", "th", "tr", "uk", "vi", "zh-CN", and "zh-TW".
    separator
        String to display between the start and end input boxes.
    width
        The CSS width, e.g. '400px', or '100%'
    autoclose
        Whether or not to close the datepicker immediately when a date is selected.

    Returns
    -------
    :
        A UI element.

    Note
    ----
    The date ``format`` string specifies how the date will be displayed in the browser.
    It allows the following values:

    - ``yy``: Year without century (12)
    - ``yyyy``: Year with century (2012)
    - ``mm``: Month number, with leading zero (01-12)
    - ``m``: Month number, without leading zero (1-12)
    - ``M``: Abbreviated month name
    - ``MM``: Full month name
    - ``dd``: Day of month with leading zero
    - ``d``: Day of month without leading zero
    - ``D``: Abbreviated weekday name
    - ``DD``: Full weekday name

    Notes
    ------
    ::: {.callout-note title="Server value"}
    A tuple of :class:`~datetime.date` objects.
    :::

    See Also
    --------
    * :func:`~shiny.ui.update_date_range`
    * :func:`~shiny.ui.input_date`
    """

    resolved_id = resolve_id(id)
    return div(
        shiny_input_label(resolved_id, label),
        div(
            _date_input_tag(
                id=resolved_id,
                value=start,
                min=min,
                max=max,
                format=format,
                startview=startview,
                weekstart=weekstart,
                language=language,
                autoclose=autoclose,
            ),
            # input-group-prepend and input-group-append are for bootstrap 4 forward compat
            span(
                span(separator, class_="input-group-text"),
                class_="input-group-addon input-group-prepend input-group-append",
            ),
            _date_input_tag(
                id=resolved_id,
                value=end,
                min=min,
                max=max,
                format=format,
                startview=startview,
                weekstart=weekstart,
                language=language,
                autoclose=autoclose,
            ),
            # input-daterange class is needed for dropdown behavior
            class_="input-daterange input-group input-group-sm",
        ),
        id=resolved_id,
        class_="shiny-date-range-input form-group shiny-input-container",
        style=css(width=width),
    )


def _date_input_tag(
    id: str,
    value: Optional[date | str],
    min: Optional[date | str],
    max: Optional[date | str],
    format: str,
    startview: str,
    weekstart: int,
    language: str,
    autoclose: bool,
    **kwargs: TagAttrValue,
):
    return tags.input(
        datepicker_deps(),
        {"class": "form-control"},
        type="text",
        # `aria-labelledby` attribute is required for accessibility to avoid doubled labels (#2951).
        aria_labelledby=id + "-label",
        # title attribute is announced for screen readers for date format.
        title="Date format: " + format,
        data_date_language=language,
        data_date_week_start=weekstart,
        data_date_format=format,
        data_date_start_view=startview,
        data_min_date=_as_date_attr(min),
        data_max_date=_as_date_attr(max),
        data_initial_date=_as_date_attr(value),
        data_date_autoclose="true" if autoclose else "false",
        **kwargs,
    )


def _as_date_attr(x: Optional[date | str]) -> Optional[str]:
    if x is None:
        return None
    if isinstance(x, date):
        return str(x)
    return str(date.fromisoformat(x))



from shiny import reactive
from shiny.express import input, ui

ui.input_switch("update_panel", "Update (and open) Sections")

with ui.accordion(id="acc", multiple=True):
    for letter in "ABCDE":
        with ui.accordion_panel(f"Section {letter}", value=f"sec_{letter}"):
            f"Some narrative for section {letter}"


@reactive.effect
@reactive.event(input.update_panel)
def _():
    txt = " (updated)" if input.update_panel() else ""
    show = bool(input.update_panel() % 2 == 1)
    for letter in "ABCDE":
        ui.update_accordion_panel(
            "acc",
            f"sec_{letter}",
            f"Some{txt} narrative for section {letter}",
            title=f"Section {letter}{txt}",
            # Open Accordion Panel to see updated contents
            show=show,
        )
    next_show_txt = "close" if show else "open"

    ui.update_switch("update_panel", label=f"Update (and {next_show_txt}) Sections")



from datetime import datetime

from shiny.express import input, render, session, ui

ui.input_action_button("flush", "Trigger flush")


@render.ui
def n_clicks():
    return "Number of clicks: " + str(input.flush())


ui.div(id="flush_time")


def log():
    msg = "A reactive flush occurred at " + datetime.now().strftime("%H:%M:%S:%f")
    print(msg)
    ui.insert_ui(
        ui.p(msg),
        selector="#flush_time",
    )


if hasattr(session, "on_flush"):
    _ = session.on_flush(log, once=False)



from __future__ import annotations

import sys
from types import TracebackType
from typing import Callable, Optional, Type, TypeVar

from htmltools import wrap_displayhook_handler

from ..._docstring import no_example
from ..._typing_extensions import ParamSpec

__all__ = ("hold",)

P = ParamSpec("P")
R = TypeVar("R")
CallableT = TypeVar("CallableT", bound=Callable[..., object])


@no_example()
def hold() -> HoldContextManager:
    """Prevent the display of UI elements in various ways.

    This is used as a context manager, as in `with hold():`. It prevents the display of
    all UI elements within the context block. (This is useful when you want to
    temporarily prevent the display of a large number of UI elements, or when you want
    to prevent the display of UI elements that are not directly under your control.)

    It can also be used as `with hold() as content:` to capture the UI elements that
    would be displayed within the context block. Then, later, you can put `content` on a
    line by itself to display the captured UI elements.

    Returns
    -------
    :
        A context manager that prevents the display of UI elements within the context
        block.

    See Also
    --------
    * ~shiny.render.express
    * ~shiny.express.expressify
    """

    return HoldContextManager()


class HoldContextManager:
    def __init__(self):
        self.content: list[object] = list()

    def __enter__(self) -> list[object]:
        self.prev_displayhook = sys.displayhook
        sys.displayhook = wrap_displayhook_handler(
            self.content.append  # pyright: ignore[reportArgumentType]
        )
        return self.content

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> bool:
        sys.displayhook = self.prev_displayhook
        return False



from playwright.sync_api import Page, expect

from shiny.run import ShinyAppProc


def test_implicit_register(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    expect(page.locator("#txt")).to_have_text("Hello")



from htmltools import Tag

from shiny import App, Inputs, Outputs, Session, module, ui


@module.ui
def chat_mod_ui() -> Tag:
    return ui.chat_ui(id="chat")


@module.server
def chat_mod_server(input: Inputs, output: Outputs, session: Session):
    chat = ui.Chat(id="chat")

    @chat.on_user_submit
    async def _():
        user = chat.user_input()
        await chat.append_message(f"You said: {user}")


app_ui = ui.page_fillable(
    ui.panel_title("Hello Shiny Chat"),
    chat_mod_ui("foo"),
    fillable_mobile=True,
)


def server(input: Inputs, output: Outputs, session: Session):
    chat_mod_server("foo")


app = App(app_ui, server)
